from plenum.common.request import Request
from plenum.common.types import f


def generate_action_result(request: Request):
    return {**request.operation, **{
        f.IDENTIFIER.nm: request.identifier,
        f.REQ_ID.nm: request.reqId}}



import pytest
import time

from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.constants import ENDORSER
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.constants import TRUSTEE
from plenum.test.helper import randomOperation
from storage.kv_in_memory import KeyValueStorageInMemory


@pytest.fixture(scope='module')
def identity_owners():
    return ["identity_owners_{}".format(i) for i in range(5)]


@pytest.fixture(scope='module')
def trustees():
    return ["trustee_{}".format(i) for i in range(5)]


@pytest.fixture(scope='module')
def endorsers():
    return ["endorser_{}".format(i) for i in range(5)]


@pytest.fixture(scope='module')
def idr_cache(identity_owners, trustees, endorsers):
    cache = IdrCache("Cache",
                     KeyValueStorageInMemory())
    seq_no = 1
    for identifier in identity_owners:
        cache.set(identifier, seq_no, int(time.time()), role="",
                  verkey="owner_identifier_verkey", isCommitted=False)

    for identifier in trustees:
        cache.set(identifier, seq_no, int(time.time()), role=TRUSTEE,
                  verkey="trustee_identifier_verkey", isCommitted=False)
    for identifier in endorsers:
        cache.set(identifier, seq_no, int(time.time()), role=ENDORSER,
                  verkey="endorser_identifier_verkey", isCommitted=False)
    return cache


@pytest.fixture(scope='module')
def req(identity_owners):
    return Request(identifier=identity_owners[0],
                   operation=randomOperation())


@pytest.fixture(scope='module')
def key():
    return AuthActionAdd(txn_type='SomeType',
                         field='some_field',
                         value='new_value')



from plenum.common.init_util import initialize_node_environment as \
    p_initialize_node_environment, cleanup_environment


def initialize_node_environment(name, node_config_helper, sigseed=None,
                                override_keep=False):
    cleanup_environment(node_config_helper.ledger_dir)
    vk, bls_key, key_proof = p_initialize_node_environment(name, node_config_helper, sigseed, override_keep)

    return vk, bls_key, key_proof



from plenum.common.constants import TARGET_NYM, TXN_TYPE
import pytest
from indy_common.constants import DIDDOC_CONTENT, NYM
from indy_common.types import ClientNYMOperation

VALID_TARGET_NYM = 'a' * 43


@pytest.fixture
def validator():
    yield ClientNYMOperation()


def test_nym(validator):
    """Validate that the NYM transaction accepts only JSON
    strings as DIDDOC_CONTENT."""

    msg = {
        TXN_TYPE: NYM,
        TARGET_NYM: VALID_TARGET_NYM,
        DIDDOC_CONTENT: '{}',
    }
    validator.validate(msg)


def test_nym_raw_dictionary(validator):
    """Validate that the NYM transaction does not accept
    dictionaries as DIDDOC_CONTENT."""
    msg = {
        TXN_TYPE: NYM,
        TARGET_NYM: VALID_TARGET_NYM,
        DIDDOC_CONTENT: {},
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)

    ex_info.match("validation error")



import json
import pytest

from plenum.common.constants import STEWARD_STRING
from plenum.common.exceptions import RequestNackedException
from plenum.common.util import randomString
from plenum.test.helper import sdk_multi_sign_request_objects, \
    sdk_json_to_request_object, sdk_send_signed_requests, sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import prepare_nym_request


def test_txn_with_different_signature_and_idr(
        looper, txnPoolNodeSet, sdk_pool_handle, sdk_wallet_trustee, sdk_wallet_client):
    # filling nym request and getting steward did
    nym_request, new_did = looper.loop.run_until_complete(
        prepare_nym_request(sdk_wallet_trustee, randomString(32),
                            'newSteward1', STEWARD_STRING))

    # sending request using 'sdk_' functions
    signed_reqs = sdk_multi_sign_request_objects(looper, [sdk_wallet_client],
                                                 [sdk_json_to_request_object(
                                                     json.loads(nym_request))])
    request_couple = sdk_send_signed_requests(sdk_pool_handle, signed_reqs)[0]

    with pytest.raises(RequestNackedException, match="The identifier is not contained in signatures"):
        # waiting for replies
        sdk_get_and_check_replies(looper, [request_couple])



from itertools import combinations, product

import pytest

from indy_common.authorize.auth_constraints import IDENTITY_OWNER
from indy_common.constants import ENDORSER
from indy_common.test.auth.metadata.helper import PluginAuthorizer
from plenum.common.constants import TRUSTEE, STEWARD
from plenum.test.conftest import getValueFromModule

ROLES = [TRUSTEE, STEWARD, ENDORSER, IDENTITY_OWNER]
MAX_SIG_COUNT = 3


@pytest.fixture(scope='module')
def write_auth_req_validator(write_auth_req_validator):
    plugin_authorizer = PluginAuthorizer()
    write_auth_req_validator.register_authorizer(plugin_authorizer)
    return write_auth_req_validator


@pytest.fixture(scope='module', params=[None, 0, 1, 2, 3, 4, 100])
def amount(request):
    return request.param


@pytest.fixture(scope='module', params=[True, False])
def is_owner(request):
    return request.param


@pytest.fixture(scope='module', params=[None, ENDORSER])
def endorser(request):
    return request.param


@pytest.fixture(scope='module', params=[IDENTITY_OWNER, TRUSTEE])
def author(request):
    return request.param


@pytest.fixture(
    scope='module',
    params=[role for r in range(len(ROLES) + 1) for role in combinations(ROLES, r)],
    ids=[str(role) for r in range(len(ROLES) + 1) for role in combinations(ROLES, r)]
)
def roles(request):
    '''
    Combination of all roles.

    Example:
      - roles=[A, B, C]
      ==> [(A), (B), (C), (A, B), (A, C), (B, C), (A, B, C)]
    '''
    return request.param


@pytest.fixture(scope='module')
def signatures(request, roles):
    '''
    Combinations of all signature types and signature count.

    Example:
          - roles=[A, B]
          - sig_count=1..3
          =>[(), (A: 1), (B:1), (A: 2), (B: 2), (A: 3), (B: 3),
             (A:1, B: 1), (A:2, B: 1), (A:1, B: 2), (A:2, B: 2),
             (A:1, B: 3), (A:3, B: 1), (A:3, B: 3),
             (A:2, B: 3), (A:3, B: 2)]

    '''
    max_sig_count = getValueFromModule(request, "MAX_SIG_COUNT", 3)
    all_sigs_count = list(range(1, max_sig_count))
    return [
        {role: sig_count for role, sig_count in zip(roles, sigs_count)}
        for sigs_count in product(all_sigs_count, repeat=len(roles))
    ]



from indy_common.constants import GET_RS_SCHEMA, RS_META, RS_META_NAME, RS_META_VERSION, RS_SCHEMA_FROM, \
    DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rs_schema_handler import RsSchemaHandler


class GetRsSchemaHandler(ReadRequestHandler):
    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, GET_RS_SCHEMA, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        author_did = request.operation[RS_SCHEMA_FROM]
        name = request.operation[RS_META][RS_META_NAME]
        version = request.operation[RS_META][RS_META_VERSION]
        state_path = RsSchemaHandler.make_state_path(author_did, name, version)

        rs_schema, last_seq_no, last_update_time, proof = {}, None, None, None
        try:  # lookup returns: keys, seq_no, last_update_time, proof
            rs_schema, last_seq_no, last_update_time, proof = self.lookup(state_path, True, True)
        except KeyError:
            pass
        return self.make_result(request=request, data=rs_schema, last_seq_no=last_seq_no, update_time=last_update_time,
                                proof=proof)



# TODO it's a copy-paste of common.version:
# remove once plenum is merged to node
import pytest

from indy_common.node_version import NodeVersion, InvalidVersionError


# valid PEP440:
#  alpha prerelease
#  beta prerelease
#  postrelease
#  epoch
#  local version
#  parts num != 3
@pytest.mark.parametrize(
    'version',
    [
        '1.2.3a1',
        '1.2.3b2',
        '1.2.3.post1',
        '1!1.2.3',
        '1.2.3+1',
        '1',
        # '1.2', TODO uncomment once new release logic becomes completed
        '1.2.3.4'
    ]
)
def test_node_version_invalid_value(version):
    with pytest.raises(InvalidVersionError):
        NodeVersion(version)


@pytest.mark.parametrize(
    'version',
    [
        '1.2.3',
        '1.2.3.rc1',
        '1.2.3.dev2',
    ]
)
def test_node_version_valid(version):
    NodeVersion(version)


def test_node_version_parts():
    assert NodeVersion('1.2.3.dev2').parts == (1, 2, 3, 'dev', 2)
    assert NodeVersion('1.2.3.rc3').parts == (1, 2, 3, 'rc', 3)
    assert NodeVersion('1.2.3').parts == (1, 2, 3, None, None)


def test_node_version_upstream():
    pv = NodeVersion('1.2.3')
    assert pv.upstream is pv



import pytest

from plenum.common.exceptions import RequestNackedException

from indy_node.test.state_proof.test_state_proof_for_get_requests import check_get_delta

from indy_node.test.anon_creds.conftest import send_revoc_reg_entry, send_revoc_reg_def, send_claim_def, claim_def
from indy_node.test.schema.test_send_get_schema import send_schema_req
from indy_node.test.state_proof.conftest import nodeSetWithOneNodeResponding


def test_state_proof_returned_for_get_revoc_reg_delta(looper,
                                                      nodeSetWithOneNodeResponding,
                                                      sdk_wallet_steward,
                                                      sdk_pool_handle,
                                                      sdk_wallet_client,
                                                      send_revoc_reg_entry):
    revoc_reg_def = send_revoc_reg_entry[0]
    revoc_reg_entry_data = send_revoc_reg_entry[1][0]['operation']
    timestamp = send_revoc_reg_entry[1][1]['result']['txnMetadata']['txnTime']

    with pytest.raises(RequestNackedException, match='Timestamp FROM more then TO'):
        check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, timestamp + 1, timestamp,
                        sdk_pool_handle, revoc_reg_entry_data)



from re import findall

from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import JSON_LD_CONTEXT, RS_CONTENT, JSON_LD_CONTEXT_FIELD
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.abstract_rich_schema_object_handler import \
    AbstractRichSchemaObjectHandler
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager

URI_REGEX = r'^(?P<scheme>\w+):(?:(?:(?P<url>//[.\w]+)(?:(/(?P<path>[/\w]+)?)?))|(?:(?P<method>\w+):(?P<id>\w+)))'


class JsonLdContextHandler(AbstractRichSchemaObjectHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(JSON_LD_CONTEXT, database_manager, write_req_validator)

    def is_json_ld_content(self):
        return False

    def do_static_validation_content(self, content_as_dict, request: Request):
        if JSON_LD_CONTEXT_FIELD not in content_as_dict:
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "'{}' must contain a {} field".format(RS_CONTENT, JSON_LD_CONTEXT_FIELD))

        self._validate_context(content_as_dict[JSON_LD_CONTEXT_FIELD], request.identifier, request.reqId)

    def do_dynamic_validation_content(self, request):
        pass

    def _validate_context(self, context, id, reqId):
        if isinstance(context, list):
            for ctx in context:
                if not isinstance(ctx, dict):
                    if self._bad_uri(ctx):
                        raise InvalidClientRequest(id, reqId, '@context URI {} badly formed'.format(ctx))
        elif isinstance(context, dict):
            pass
        elif isinstance(context, str):
            if self._bad_uri(context):
                raise InvalidClientRequest(id, reqId, '@context URI {} badly formed'.format(context))
        else:
            raise InvalidClientRequest(id, reqId, "'@context' value must be url, array, or object")

    def _bad_uri(self, uri_string):
        url = findall(URI_REGEX, uri_string)
        if not url:
            return True
        return False



import os
import shutil
import subprocess

from indy_common.util import compose_cmd


def rename_if_exists(dir, old_name, new_name):
    if os.path.exists(os.path.join(dir, old_name)):
        os.rename(os.path.join(dir, old_name),
                  os.path.join(dir, new_name))


def rename_request_files(requests_dir):
    for relative_name in os.listdir(requests_dir):
        absolute_name = os.path.join(requests_dir, relative_name)
        if os.path.isfile(absolute_name) and absolute_name.endswith('.sovrin'):
            os.rename(absolute_name, absolute_name[:-len('.sovrin')] + '.indy')


def migrate():
    source_dir = os.path.expanduser('/home/sovrin/.sovrin')
    target_dir = os.path.expanduser('/home/indy/.indy')

    if os.path.isdir(target_dir):
        shutil.rmtree(target_dir)
    shutil.copytree(source_dir, target_dir)

    rename_if_exists(target_dir, '.sovrin', '.indy')
    rename_if_exists(target_dir, 'sovrin.env', 'indy.env')
    rename_if_exists(target_dir, 'sovrin_config.py', 'indy_config.py')

    if os.path.isdir(os.path.join(target_dir, 'sample')):
        rename_request_files(os.path.join(target_dir, 'sample'))

    subprocess.run(compose_cmd(['chown', '-R', 'indy:indy', target_dir]),
                   shell=True,
                   check=True)


migrate()



import json

from indy.did import create_and_store_my_did

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.constants import AUTH_RULE, CONSTRAINT, AUTH_ACTION, AUTH_TYPE, FIELD, NEW_VALUE, OLD_VALUE, NYM, \
    ENDORSER, RULES, AUTH_RULES, GET_AUTH_RULE
from plenum.common.constants import TRUSTEE, TXN_TYPE

from indy_common.authorize.auth_constraints import CONSTRAINT_ID, ROLE, SIG_COUNT, NEED_TO_BE_OWNER, METADATA, \
    ConstraintsEnum, AUTH_CONSTRAINTS
from plenum.common.util import randomString

from plenum.test.helper import sdk_sign_and_submit_req_obj, sdk_get_and_check_replies, sdk_gen_request, \
    sdk_multi_sign_request_objects, sdk_json_to_request_object, sdk_send_signed_requests
from plenum.test.pool_transactions.helper import prepare_nym_request

from indy_node.test.helper import (
    sdk_send_and_check_req_json,
    sdk_send_and_check_auth_rule_request as _sdk_send_and_check_auth_rule_request,
    sdk_send_and_check_get_auth_rule_request as _sdk_send_and_check_get_auth_rule_request,
    generate_auth_rule, generate_constraint_entity)


def generate_constraint_list(constraint_id=ConstraintsEnum.AND_CONSTRAINT_ID,
                             auth_constraints=None):
    auth_constraints = generate_constraint_entity() \
        if auth_constraints is None \
        else auth_constraints
    return {CONSTRAINT_ID: constraint_id,
            AUTH_CONSTRAINTS: auth_constraints}


# Note. The helper is for testing invalid operations/requests only.
# DO NOT USE it for valid cases, use 'build_auth_rule_request_json'
# instead.
def generate_auth_rule_operation(auth_action=ADD_PREFIX, auth_type=NYM,
                                 field=ROLE, new_value=ENDORSER,
                                 old_value=None, constraint=None):
    op = generate_auth_rule(auth_action, auth_type,
                            field, new_value,
                            old_value, constraint)
    op[TXN_TYPE] = AUTH_RULE
    return op


def create_verkey_did(looper, wh):
    seed = randomString(32)
    return looper.loop.run_until_complete(
        create_and_store_my_did(wh, json.dumps({'seed': seed})))


def sdk_send_and_check_auth_rule_request(
    looper, sdk_pool_handle, sdk_wallet_trustee,
    auth_action=ADD_PREFIX,
    auth_type=NYM,
    field=ROLE,
    old_value=None,
    new_value=ENDORSER,
    constraint=None,
    no_wait=False
):
    constraint = (
        generate_constraint_entity() if constraint is None else constraint
    )
    return _sdk_send_and_check_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee,
        auth_action=auth_action,
        auth_type=auth_type,
        field=field,
        old_value=old_value,
        new_value=new_value,
        constraint=constraint,
        no_wait=no_wait
    )


def sdk_send_and_check_get_auth_rule_request(
    looper, sdk_pool_handle, sdk_wallet,
    auth_type=None,
    auth_action=None,
    field=None,
    old_value=None,
    new_value=None
):
    return _sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet,
        auth_action=auth_action,
        auth_type=auth_type,
        field=field,
        old_value=old_value,
        new_value=new_value
    )


def sdk_send_and_check_get_auth_rule_invalid_request(
    looper, sdk_pool_handle, sdk_wallet, **invalid_params
):
    op = {TXN_TYPE: GET_AUTH_RULE}
    op.update(**invalid_params)
    req_obj = sdk_gen_request(op, identifier=sdk_wallet[1])
    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, sdk_wallet, json.dumps(req_obj.as_dict)
    )


def sdk_send_and_check_auth_rule_invalid_request(
    looper, sdk_pool_handle, sdk_wallet_trustee, no_wait=False, **invalid_params
):
    op = generate_auth_rule_operation(**invalid_params)
    req_obj = sdk_gen_request(op, identifier=sdk_wallet_trustee[1])
    req_json = json.dumps(req_obj.as_dict)
    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, sdk_wallet_trustee, req_json, no_wait=no_wait
    )


def sdk_send_and_check_auth_rules_request_invalid(looper, sdk_pool_handle, sdk_wallet_trustee,
                                          rules=None, no_wait=False):
    if rules is None:
        rules = [generate_auth_rule(ADD_PREFIX, NYM,
                                    ROLE, ENDORSER),
                 generate_auth_rule(EDIT_PREFIX, NYM,
                                    ROLE, ENDORSER, TRUSTEE)]
    op = {RULES: rules,
          TXN_TYPE: AUTH_RULES}

    req_obj = sdk_gen_request(op, identifier=sdk_wallet_trustee[1])
    req = sdk_sign_and_submit_req_obj(looper,
                                      sdk_pool_handle,
                                      sdk_wallet_trustee,
                                      req_obj)
    if no_wait:
        return req
    resp = sdk_get_and_check_replies(looper, [req])
    return resp


def add_new_nym(looper, sdk_pool_handle, creators_wallets,
                alias=None, role=None, seed=None,
                dest=None, verkey=None, skipverkey=False, no_wait=False):
    seed = seed or randomString(32)
    alias = alias or randomString(5)
    wh, _ = creators_wallets[0]

    # filling nym request and getting steward did
    # if role == None, we are adding client
    nym_request, new_did = looper.loop.run_until_complete(
        prepare_nym_request(creators_wallets[0], seed,
                            alias, role, dest, verkey, skipverkey))

    # sending request using 'sdk_' functions
    signed_reqs = sdk_multi_sign_request_objects(looper, creators_wallets,
                                                 [sdk_json_to_request_object(
                                                     json.loads(nym_request))])
    request_couple = sdk_send_signed_requests(sdk_pool_handle, signed_reqs)[0]
    if no_wait:
        return request_couple
    # waiting for replies
    sdk_get_and_check_replies(looper, [request_couple])


def generate_key(auth_action=ADD_PREFIX, auth_type=NYM,
                 field=ROLE, new_value=ENDORSER,
                 old_value=None):
    key = {AUTH_ACTION: auth_action,
           AUTH_TYPE: auth_type,
           FIELD: field,
           NEW_VALUE: new_value,
           }
    if old_value:
        key[OLD_VALUE] = old_value
    return key


def get_pool_validator_count(pool):
    vsb = set(len(n.nodeReg) for n in pool)
    assert len(vsb) == 1
    return vsb.pop()



import pytest
from indy_common.authorize.auth_actions import EDIT_PREFIX, ADD_PREFIX

from indy_common.constants import OLD_VALUE, AUTH_TYPE, ROLE, ENDORSER, AUTH_ACTION, \
    AUTH_RULES, RULES
from indy_node.server.request_handlers.config_req_handlers.auth_rule.abstract_auth_rule_handler import \
    AbstractAuthRuleHandler

from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rules_handler import AuthRulesHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from indy_node.test.auth_rule.helper import generate_auth_rule
from indy_node.test.request_handlers.helper import add_to_idr
from plenum.common.constants import TRUSTEE, TXN_TYPE, STEWARD
from plenum.common.exceptions import InvalidClientRequest, UnauthorizedClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import reqToTxn, get_payload_data


@pytest.fixture(scope="module")
def auth_rules_handler(db_manager, write_auth_req_validator):
    return AuthRulesHandler(db_manager, write_auth_req_validator)


@pytest.fixture(scope="function")
def auth_rules_request(creator):
    return Request(identifier=creator,
                   reqId=5,
                   operation={TXN_TYPE: AUTH_RULES,
                              RULES: [generate_auth_rule()]})


def test_auth_rule_static_validation(auth_rules_request, auth_rules_handler: AuthRulesHandler):
    auth_rules_handler.static_validation(auth_rules_request)


def test_auth_rule_static_validation_failed_without_old_value(auth_rules_request,
                                                              auth_rules_handler: AuthRulesHandler):
    rule = auth_rules_request.operation[RULES][0]
    if OLD_VALUE in rule:
        del rule[OLD_VALUE]
    rule[AUTH_ACTION] = EDIT_PREFIX
    with pytest.raises(InvalidClientRequest, match="EDIT must contain field old_value"):
        auth_rules_handler.static_validation(auth_rules_request)


def test_auth_rule_static_validation_failed_with_excess_field(auth_rules_request,
                                                              auth_rules_handler: AuthRulesHandler):
    rule = auth_rules_request.operation[RULES][0]
    rule[OLD_VALUE] = "old_value"
    rule[AUTH_TYPE] = ADD_PREFIX
    with pytest.raises(InvalidClientRequest, match="ADD must not contain field old_value"):
        auth_rules_handler.static_validation(auth_rules_request)


def test_auth_rule_static_validation_failed_with_incorrect_key(auth_rules_request,
                                                               auth_rules_handler: AuthRulesHandler):
    auth_rules_request.operation[RULES].append(generate_auth_rule(auth_action=ADD_PREFIX,
                                                                 auth_type="wrong_type",
                                                                 field=ROLE, new_value=ENDORSER))
    with pytest.raises(InvalidClientRequest, match="key .* is not found in authorization map"):
        auth_rules_handler.static_validation(auth_rules_request)


def test_auth_rule_dynamic_validation_without_permission(auth_rules_request,
                                                         auth_rules_handler: AuthRulesHandler, creator):
    add_to_idr(auth_rules_handler.database_manager.idr_cache, creator, STEWARD)
    with pytest.raises(UnauthorizedClientRequest):
        auth_rules_handler.dynamic_validation(auth_rules_request, 0)


def test_auth_rule_dynamic_validation(auth_rules_request,
                                      auth_rules_handler: AuthRulesHandler, creator):
    add_to_idr(auth_rules_handler.database_manager.idr_cache, creator, TRUSTEE)


def test_update_state(auth_rules_request, auth_rules_handler: AuthRulesHandler):
    txn = reqToTxn(auth_rules_request)
    payload = get_payload_data(txn)
    for rule in payload[RULES]:
        constraint = StaticAuthRuleHelper.get_auth_constraint(rule)
        auth_key = StaticAuthRuleHelper.get_auth_key(rule)
        path = AbstractAuthRuleHandler.make_state_path_for_auth_rule(auth_key)

        auth_rules_handler.update_state(txn, None, auth_rules_request)
        assert auth_rules_handler.get_from_state(path) == constraint.as_dict



import json
from _sha256 import sha256
from contextlib import contextmanager
import pytest
from common.serializers.serialization import serialize_msg_for_signing
from indy_common.authorize.auth_actions import EDIT_PREFIX, ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_node.test.attrib_txn.test_send_get_attr import secretBox
from indy_node.test.helper import sdk_send_and_check_auth_rule_request

from plenum.common.constants import STEWARD_STRING, STEWARD
from plenum.common.exceptions import RequestRejectedException
from plenum.common.util import adict, randomString
from indy_common.constants import ENDORSER_STRING, ATTRIB
from indy_common.util import getSymmetricallyEncryptedVal
from indy_node.test.helper import sdk_add_attribute_and_check, sdk_get_attribute_and_check
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from stp_core.common.log import getlogger

logger = getlogger()


@pytest.fixture(scope="module")
def attributeName():
    return 'endpoint'


@pytest.fixture(scope="module")
def attributeValue():
    return {
        "ha": "127.0.0.1:9700",
        "verkey": "F46i9NmUN72QMbbm5qWetB6CmfT7hiU8BM1qrtTGLKsc"
    }


@pytest.fixture(scope="module")
def attributeData(attributeName, attributeValue):
    return json.dumps({attributeName: attributeValue})


@pytest.fixture(scope="module")
def sdk_added_raw_attribute(sdk_pool_handle, sdk_user_wallet_a,
                            sdk_wallet_endorser, attributeData, looper):
    _, did_cl = sdk_user_wallet_a
    req_couple = sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, attributeData, did_cl)[0]
    return req_couple[1]


@pytest.fixture(scope="module")
def symEncData(attributeData):
    encData, secretKey = getSymmetricallyEncryptedVal(attributeData)
    return adict(data=attributeData, encData=encData, secretKey=secretKey)


@contextmanager
def whitelistextras(*msg):
    global whitelistArray
    ins = {m: (m in whitelistArray) for m in msg}
    [whitelistArray.append(m) for m, _in in ins.items() if not _in]
    yield
    [whitelistArray.remove(m) for m, _in in ins.items() if not _in]


def testEndorserAddsAttributeForUser(sdk_added_raw_attribute):
    pass


def testEndorserGetAttrsForUser(looper,
                                   sdk_user_wallet_a,
                                   sdk_wallet_endorser,
                                   sdk_pool_handle,
                                   attributeName,
                                   sdk_added_raw_attribute):
    _, dest = sdk_user_wallet_a
    sdk_get_attribute_and_check(looper, sdk_pool_handle,
                                sdk_wallet_endorser, dest, attributeName)


def test_edit_attrib(sdk_pool_handle, sdk_user_wallet_a,
                     sdk_wallet_endorser, attributeData, looper, attributeName):
    _, did_cl = sdk_user_wallet_a

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, attributeData, did_cl)
    res1 = sdk_get_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, did_cl, attributeName)
    assert serialize_msg_for_signing(
        json.loads(res1[0][1]['result']['data'])) == serialize_msg_for_signing(
        json.loads(attributeData.replace(' ', '')))

    data = json.loads(attributeData)
    data[attributeName] = {'John': 'Snow'}
    data = json.dumps(data)

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)
    res2 = sdk_get_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, did_cl, attributeName)
    assert serialize_msg_for_signing(
        json.loads(res2[0][1]['result']['data'])) == serialize_msg_for_signing(
        json.loads(data.replace(' ', '')))


def test_non_endorser_cannot_add_attribute_for_user(
        looper,
        nodeSet,
        sdk_wallet_client,
        sdk_pool_handle,
        sdk_user_wallet_a,
        attributeData):
    _, dest = sdk_user_wallet_a

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle,
                                    sdk_wallet_client, attributeData, dest)
    e.match('can not touch raw field since only the owner can modify it')


def testOnlyUsersEndorserCanAddAttribute(
        nodeSet,
        looper,
        attributeData,
        sdk_pool_handle,
        sdk_wallet_trustee,
        sdk_user_wallet_a):
    _, dest = sdk_user_wallet_a
    wallet_another_ta = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                                        alias='TA-' + randomString(5), role=ENDORSER_STRING)
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle,
                                    wallet_another_ta, attributeData, dest)
    e.match('can not touch raw field since only the owner can modify it')


def testStewardCannotAddUsersAttribute(
        nodeSet,
        looper,
        attributeData,
        sdk_pool_handle,
        sdk_wallet_trustee,
        sdk_user_wallet_a):
    _, dest = sdk_user_wallet_a
    wallet_another_stewatd = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                                             alias='TA-' + randomString(5), role=STEWARD_STRING)
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle,
                                    wallet_another_stewatd, attributeData, dest)
    e.match('can not touch raw field since only the owner can modify it')


# TODO: Ask Jason, if getting the latest attribute makes sense since in case
# of encrypted and hashed attributes, there is no name.
def testLatestAttrIsReceived(
        looper,
        nodeSet,
        sdk_wallet_endorser,
        sdk_pool_handle,
        sdk_user_wallet_a):
    _, dest = sdk_user_wallet_a

    attr1 = json.dumps({'name': 'Mario'})
    sdk_add_attribute_and_check(looper, sdk_pool_handle,
                                sdk_wallet_endorser, attr1, dest)
    reply = sdk_get_attribute_and_check(looper, sdk_pool_handle,
                                        sdk_wallet_endorser, dest, 'name')[0]
    reply_equality_of_get_attribute(reply, 'Mario')

    attr2 = json.dumps({'name': 'Luigi'})
    sdk_add_attribute_and_check(looper, sdk_pool_handle,
                                sdk_wallet_endorser, attr2, dest)
    reply = sdk_get_attribute_and_check(looper, sdk_pool_handle,
                                        sdk_wallet_endorser, dest, 'name')[0]
    reply_equality_of_get_attribute(reply, 'Luigi')


def reply_equality_of_get_attribute(reply, value):
    result = reply[1]['result']
    assert json.loads(result['data'])[result['raw']] == value


def test_user_add_attrs_for_herself_and_get_it(
        looper,
        nodeSet,
        sdk_wallet_trustee,
        sdk_pool_handle):
    wallet_client = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee, role=None)
    _, dest = wallet_client
    attr = json.dumps({'name': 'Albus'})
    sdk_add_attribute_and_check(looper, sdk_pool_handle,
                                wallet_client, attr, dest)
    sdk_get_attribute_and_check(looper, sdk_pool_handle,
                                wallet_client, dest, 'name')


@pytest.mark.skip(reason="INDY-896 ATTR cannot be added without dest")
def test_attr_with_no_dest_added(nodeSet, looper, attributeData):
    pass
    # user_wallet = Wallet()
    # signer = DidSigner()
    # user_wallet.addIdentifier(signer=signer)
    #
    # client.registerObserver(user_wallet.handleIncomingReply)
    # looper.add(client)
    # looper.run(client.ensureConnectedToNodes())
    # makePendingTxnsRequest(client, user_wallet)
    #
    # createNym(looper,
    #           user_wallet.defaultId,
    #           endorser,
    #           addedEndorser,
    #           role=None,
    #           verkey=user_wallet.getVerkey())
    #
    # attr1 = json.dumps({'age': "24"})
    # attrib = Attribute(name='test4 attribute',
    #                    origin=user_wallet.defaultId,
    #                    value=attr1,
    #                    dest=None,
    #                    ledgerStore=LedgerStore.RAW)
    # addAttributeAndCheck(looper, client, user_wallet, attrib)


@pytest.mark.skip(reason="SOV-561. Test not implemented")
def testGetTxnsNoSeqNo():
    """
    Test GET_TXNS from client and do not provide any seqNo to fetch from
    """
    raise NotImplementedError


@pytest.mark.skip(reason="SOV-560. Come back to it later since "
                         "requestPendingTxns move to wallet")
def testGetTxnsSeqNo(nodeSet, endorserWallet, looper):
    pass
    """
    Test GET_TXNS from client and provide seqNo to fetch from
    """
    # looper.add(endorser)
    # looper.run(endorser.ensureConnectedToNodes())
    #
    # def chk():
    #     assert endorser.spylog.count(
    #         endorser.requestPendingTxns.__name__) > 0
    #
    # # TODO choose or create timeout in 'waits' on this case.
    # looper.run(eventually(chk, retryWait=1, timeout=3))


@pytest.mark.skip(reason="SOV-560. Attribute encryption is done in client")
def testEndorserAddedAttributeIsEncrypted(addedEncryptedAttribute):
    pass


@pytest.mark.skip(reason="SOV-560. Attribute Disclosure is not done for now")
def testEndorserDisclosesEncryptedAttribute(
        addedEncryptedAttribute,
        symEncData,
        looper,
        userSignerA,
        endorserSigner,
        endorser):
    pass
    # box = libnacl.public.Box(endorserSigner.naclSigner.keyraw,
    #                          userSignerA.naclSigner.verraw)
    #
    # data = json.dumps({SKEY: symEncData.secretKey,
    #                    TXN_ID: addedEncryptedAttribute[TXN_ID]})
    # nonce, boxedMsg = box.encrypt(data.encode(), pack_nonce=False)
    #
    # op = {
    #     TARGET_NYM: userSignerA.verstr,
    #     TXN_TYPE: ATTRIB,
    #     NONCE: base58.b58encode(nonce).decode("utf-8"),
    #     ENC: base58.b58encode(boxedMsg).decode("utf-8")
    # }
    # submitAndCheck(looper, endorser, op,
    #                identifier=endorserSigner.verstr)


@pytest.mark.skip(reason="SOV-561. Pending implementation")
def testEndorserAddedAttributeCanBeChanged(sdk_added_raw_attribute):
    # TODO but only by user(if user has taken control of his identity) and
    # endorser
    raise NotImplementedError


def set_attrib_auth_to_none(looper, sdk_wallet_trustee, sdk_pool_handle):
    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         constraint=AuthConstraint(role='*', sig_count=1,
                                                                   need_to_be_owner=True).as_dict)

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         old_value='*',
                                         constraint=AuthConstraint(role='*', sig_count=1,
                                                                   need_to_be_owner=True).as_dict)


def test_auth_rule_for_raw_attrib_works(looper,
                                        sdk_wallet_trustee,
                                        sdk_pool_handle,
                                        sdk_user_wallet_a,
                                        sdk_wallet_endorser):
    _, did_cl = sdk_user_wallet_a

    # We can add and modify attribs
    data = dict()
    data['a'] = {'John': 'Snow'}
    data = json.dumps(data)
    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         old_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=1).as_dict)

    # We still can add, but cannot edit attrib
    data = dict()
    data['b'] = {'John': 'Snow'}
    data = json.dumps(data)
    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)
    e.match('Not enough STEWARD signatures')

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=1).as_dict)
    # We cannot add or edit attrib
    data = dict()
    data['c'] = {'John': 'Snow'}
    data = json.dumps(data)
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)
    e.match('Not enough STEWARD signatures')

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, data, did_cl)
    e.match('Not enough STEWARD signatures')


def test_auth_rule_for_hash_attrib_works(looper,
                                         sdk_wallet_trustee,
                                         sdk_pool_handle,
                                         sdk_user_wallet_a,
                                         sdk_wallet_endorser):
    _, did_cl = sdk_user_wallet_a

    set_attrib_auth_to_none(looper, sdk_wallet_trustee, sdk_pool_handle)

    # We can add and modify attribs
    data = sha256(json.dumps({'name': 'John'}).encode()).hexdigest()
    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, xhash=data)

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, xhash=data)

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         old_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=1).as_dict)

    # We still can add, but cannot edit attrib
    data = sha256(json.dumps({'name': 'Ned'}).encode()).hexdigest()

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, xhash=data)

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, xhash=data)
    e.match('Not enough STEWARD signatures')

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=1).as_dict)
    # We cannot add or edit attrib
    data = sha256(json.dumps({'name': 'Aria'}).encode()).hexdigest()

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, xhash=data)
    e.match('Not enough STEWARD signatures')

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, xhash=data)
    e.match('Not enough STEWARD signatures')


def test_auth_rule_for_enc_attrib_works(looper,
                                        sdk_wallet_trustee,
                                        sdk_pool_handle,
                                        sdk_user_wallet_a,
                                        sdk_wallet_endorser):
    _, did_cl = sdk_user_wallet_a

    set_attrib_auth_to_none(looper, sdk_wallet_trustee, sdk_pool_handle)

    # We can add and modify attribs
    data = secretBox.encrypt(json.dumps({'name': 'Jaime'}).encode()).hex()
    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, enc=data)

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, enc=data)

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         old_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=1).as_dict)

    # We still can add, but cannot edit attrib
    data = secretBox.encrypt(json.dumps({'name': 'Cersei'}).encode()).hex()

    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, enc=data)

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, enc=data)
    e.match('Not enough STEWARD signatures')

    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=ATTRIB,
                                         field='*',
                                         new_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=1).as_dict)
    # We cannot add or edit attrib
    data = secretBox.encrypt(json.dumps({'name': 'Tywin'}).encode()).hex()

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, enc=data)
    e.match('Not enough STEWARD signatures')

    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser, None, did_cl, enc=data)
    e.match('Not enough STEWARD signatures')



from copy import deepcopy

from datetime import datetime, timedelta

import dateutil
import dateutil.tz

from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent


def test_pool_upgrade_same_time_different_days(looper, tconf, nodeSet,
                                               validUpgrade, sdk_pool_handle,
                                               sdk_wallet_trustee, nodeIds):
    day_in_sec = 24 * 60 * 60
    upgr1 = deepcopy(validUpgrade)
    schedule = {}
    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    startAt = unow + timedelta(seconds=day_in_sec)
    for i in nodeIds:
        schedule[i] = datetime.isoformat(startAt)
        startAt = startAt + timedelta(seconds=day_in_sec)
    upgr1['schedule'] = schedule

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr1)



import copy

import pytest
from indy_node.test.mock import build_nym_request
from plenum.common.exceptions import RequestNackedException
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request


def test_diddoc_content_added(
    looper, sdk_pool_handle, sdk_wallet_endorser, diddoc_content
):
    _, did = sdk_wallet_endorser
    nym_request = build_nym_request(did, did, None, diddoc_content, None)
    request_couple = sdk_sign_and_send_prepared_request(
        looper, sdk_wallet_endorser, sdk_pool_handle, nym_request
    )
    sdk_get_and_check_replies(looper, [request_couple])


def test_add_didoc_with_id_fails(
    looper, sdk_pool_handle, sdk_wallet_endorser, diddoc_content
):
    _, did = sdk_wallet_endorser
    diddoc_content_with_id = copy.deepcopy(diddoc_content)
    diddoc_content_with_id["id"] = "someId"
    nym_request = build_nym_request(did, did, None, diddoc_content_with_id, None)
    request_couple = sdk_sign_and_send_prepared_request(
        looper, sdk_wallet_endorser, sdk_pool_handle, nym_request
    )
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [request_couple])
    e.match("InvalidClientRequest")
    e.match("diddocContent must not have `id` at root")



from abc import abstractmethod

from indy_common.authorize.auth_cons_strategies import LocalAuthStrategy, ConfigLedgerAuthStrategy
from indy_common.authorize.auth_actions import AbstractAuthAction
from indy_common.authorize.auth_constraints import ConstraintsEnum, AbstractConstraintSerializer
from indy_common.authorize.authorizer import AbstractAuthorizer, CompositeAuthorizer, RolesAuthorizer, AndAuthorizer, \
    OrAuthorizer, AuthValidationError, ForbiddenAuthorizer, EndorserAuthorizer
from indy_common.constants import LOCAL_AUTH_POLICY, CONFIG_LEDGER_AUTH_POLICY
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.common.metrics_collector import MetricsCollector
from state.pruning_state import PruningState
from stp_core.common.log import getlogger

logger = getlogger()


class AbstractRequestValidator(AbstractAuthorizer):

    @abstractmethod
    def validate(self, request, action_list: [AbstractAuthAction]):
        raise NotImplementedError()


class WriteRequestValidator(AbstractRequestValidator, CompositeAuthorizer):
    def __init__(self,
                 config,
                 auth_map: dict,
                 cache: IdrCache,
                 config_state: PruningState,
                 state_serializer: AbstractConstraintSerializer,
                 metrics: MetricsCollector = None):
        CompositeAuthorizer.__init__(self)
        self.config = config
        self.auth_map = auth_map
        self.cache = cache
        self.config_state = config_state
        self.state_serializer = state_serializer
        self.metrics = metrics

        self.auth_cons_strategy = self.create_auth_strategy()
        self.register_default_authorizers()

    def register_default_authorizers(self):
        self.register_authorizer(RolesAuthorizer(cache=self.cache),
                                 auth_constraint_id=ConstraintsEnum.ROLE_CONSTRAINT_ID)
        self.register_authorizer(EndorserAuthorizer(cache=self.cache),
                                 auth_constraint_id=ConstraintsEnum.ROLE_CONSTRAINT_ID)
        self.register_authorizer(AndAuthorizer(), auth_constraint_id=ConstraintsEnum.AND_CONSTRAINT_ID)
        self.register_authorizer(OrAuthorizer(), auth_constraint_id=ConstraintsEnum.OR_CONSTRAINT_ID)
        self.register_authorizer(ForbiddenAuthorizer(), auth_constraint_id=ConstraintsEnum.FORBIDDEN_CONSTRAINT_ID)

    def validate(self, request: Request, action_list: [AbstractAuthAction]):
        for action in action_list:
            action_id = action.get_action_id()
            auth_constraint = self.auth_cons_strategy.get_auth_constraint(action_id)
            if auth_constraint:
                try:
                    super().authorize(request=request,
                                      auth_constraint=auth_constraint,
                                      auth_action=action)
                except AuthValidationError as exp:
                    logger.warning("Request {} cannot be authorized by reason: {}".format(request, exp.reason))
                    raise UnauthorizedClientRequest(request.identifier,
                                                    request.reqId,
                                                    exp.reason)
                return True
            error_msg = 'Request can not be authorized as action is not allowed: {}'.format(request)
            logger.warning(error_msg)
            raise UnauthorizedClientRequest(request.identifier,
                                            request.reqId,
                                            error_msg)

    def create_auth_strategy(self):
        """depends on config"""
        if self.config.authPolicy == LOCAL_AUTH_POLICY:
            return LocalAuthStrategy(auth_map=self.auth_map)
        elif self.config.authPolicy == CONFIG_LEDGER_AUTH_POLICY:
            return ConfigLedgerAuthStrategy(auth_map=self.auth_map,
                                            state=self.config_state,
                                            serializer=self.state_serializer,
                                            metrics=self.metrics)



import random

from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.common.txn_util import get_seq_no
from plenum.common.util import randomString
from plenum.test.testing_utils import FakeSomething


def add_to_idr(idr, identifier, role):
    random_s = randomString()
    idr.set(identifier,
            seqNo=5,
            txnTime=random.randint(10, 100000),
            ta=random_s,
            role=role,
            verkey=random_s,
            isCommitted=True)


def get_fake_ledger():
    ledger = FakeSomething()
    ledger.txn_list = {}
    ledger.getBySeqNo = lambda seq_no: ledger.txn_list[seq_no]
    ledger.appendTxns = lambda txns: ledger.txn_list.update({get_seq_no(txn): txn
                                                             for txn in txns})
    ledger.get_by_seq_no_uncommitted = lambda seq_no: ledger.txn_list[seq_no]
    return ledger


def get_exception(is_exception):
    def exception(request, action_list):
        if is_exception:
            raise UnauthorizedClientRequest(None, None)
        else:
            pass

    return exception



import json

from indy.ledger import build_nym_request, sign_request, submit_request

from indy_common.state import domain
from indy_node.test.helper import start_stopped_node, createHalfKeyIdentifierAndAbbrevVerkey
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.txn_util import get_txn_time
from plenum.test.node_catchup.helper import waitNodeDataEquality
from plenum.test.pool_transactions.helper import disconnect_node_and_ensure_disconnected


def test_idr_cache_update_after_catchup(txnPoolNodeSet,
                                        looper,
                                        sdk_pool_handle,
                                        sdk_wallet_steward,
                                        tconf,
                                        tdir,
                                        allPluginsPath):
    wallet_handle, identifier = sdk_wallet_steward
    node_to_disconnect = txnPoolNodeSet[-1]
    disconnect_node_and_ensure_disconnected(looper,
                                            txnPoolNodeSet,
                                            node_to_disconnect.name,
                                            stopNode=True)
    looper.removeProdable(node_to_disconnect)

    idr, verkey = createHalfKeyIdentifierAndAbbrevVerkey()
    request = looper.loop.run_until_complete(build_nym_request(identifier, idr, verkey, None, None))
    req_signed = looper.loop.run_until_complete(sign_request(wallet_handle, identifier, request))
    result = json.loads(looper.loop.run_until_complete(submit_request(sdk_pool_handle, req_signed)))

    restarted_node = start_stopped_node(node_to_disconnect, looper,
                                        tconf, tdir, allPluginsPath)
    txnPoolNodeSet[-1] = restarted_node
    waitNodeDataEquality(looper, restarted_node, *txnPoolNodeSet[:-1])
    root_hash = restarted_node.db_manager.ts_store.get_equal_or_prev(get_txn_time(result['result']))
    key = domain.make_state_path_for_nym(idr)
    from_state = restarted_node.getState(DOMAIN_LEDGER_ID).get_for_root_hash(root_hash=root_hash,
                                                                                   key=key)
    assert from_state
    deserialized = restarted_node.write_manager.state_serializer.deserialize(from_state)
    assert deserialized
    items_after = restarted_node.db_manager.idr_cache.get(idr)
    assert items_after



import os
import select
import shutil
import socket
from typing import List

from common.version import InvalidVersionError

from indy_common.constants import UPGRADE_MESSAGE, RESTART_MESSAGE, MESSAGE_TYPE
from stp_core.common.log import getlogger, Logger

from indy_common.version import (
    PackageVersion, SourceVersion, src_version_cls
)
from indy_common.config_util import getConfig
from indy_common.config_helper import ConfigHelper
from indy_common.util import compose_cmd
from indy_node.utils.migration_tool import migrate
from indy_node.utils.node_control_utils import NodeControlUtil


TIMEOUT = 600
BACKUP_FORMAT = 'zip'
BACKUP_NUM = 10
TMP_DIR = '/tmp/.indy_tmp'
LOG_FILE_NAME = 'node_control.log'


logger = getlogger()


class NodeControlTool:
    MAX_LINE_SIZE = 1024

    def __init__(
            self,
            timeout: int = TIMEOUT,
            backup_format: str = BACKUP_FORMAT,
            test_mode: bool = False,
            backup_target: str = None,
            files_to_preserve: List[str] = None,
            backup_dir: str = None,
            backup_name_prefix: str = None,
            backup_num: int = BACKUP_NUM,
            hold_ext: str = '',
            config=None):
        self.config = config or getConfig()

        self.test_mode = test_mode
        self.timeout = timeout or TIMEOUT

        self.hold_ext = hold_ext.split(" ")

        config_helper = ConfigHelper(self.config)
        self.backup_dir = backup_dir or config_helper.backup_dir
        self.backup_target = backup_target or config_helper.genesis_dir

        self.tmp_dir = TMP_DIR
        self.backup_format = backup_format

        _files_to_preserve = [self.config.lastRunVersionFile, self.config.nextVersionFile,
                              self.config.upgradeLogFile, self.config.lastVersionFilePath,
                              self.config.restartLogFile]

        self.files_to_preserve = files_to_preserve or _files_to_preserve
        self.backup_num = backup_num

        _backup_name_prefix = '{}_backup_'.format(self.config.NETWORK_NAME)

        self.backup_name_prefix = backup_name_prefix or _backup_name_prefix
        self._enable_file_logging()

        self._listen()

    def _enable_file_logging(self):
        path_to_log_file = os.path.join(self.config.LOG_DIR, LOG_FILE_NAME)
        Logger().enableFileLogging(path_to_log_file)

    def _listen(self):
        # Create a TCP/IP socket
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.setblocking(0)

        # Bind the socket to the port
        self.server_address = ('localhost', 30003)

        logger.info('Node control tool is starting up on {} port {}'.format(*self.server_address))
        self.server.bind(self.server_address)

        # Listen for incoming connections
        self.server.listen(1)

    def _get_deps_list(self, package):
        # We assume, that package looks like 'package_name=1.2.3'
        logger.info('Getting dependencies for {}'.format(package))
        NodeControlUtil.update_package_cache()
        holded = NodeControlUtil.get_sys_holds()
        if not holded:
            return package

        app_holded = list(set(self.config.PACKAGES_TO_HOLD + self.hold_ext + holded))
        deps_list = NodeControlUtil.get_deps_tree_filtered(package, hold_list=app_holded, deps_map={})
        # we need to make sure, that all hold package should be presented
        # in result list
        if package not in deps_list:
            deps_list.append(package)

        return " ".join(deps_list)

    def _call_upgrade_script(self, pkg_name: str, pkg_ver: PackageVersion):
        logger.info(
            "Upgrading {} to package version {}, test_mode {}"
            .format(pkg_name, pkg_ver, int(self.test_mode))
        )

        deps = self._get_deps_list('{}={}'.format(pkg_name, pkg_ver))
        deps = '"{}"'.format(deps)

        cmd_file = 'upgrade_indy_node'
        if self.test_mode:
            cmd_file = 'upgrade_indy_node_test'

        cmd = compose_cmd([cmd_file, deps])
        NodeControlUtil.run_shell_script(cmd, timeout=self.timeout)

    def _call_restart_node_script(self):
        logger.info('Restarting indy')
        cmd = compose_cmd(['restart_indy_node'])
        NodeControlUtil.run_shell_script(cmd, timeout=self.timeout)

    def _backup_name(self, src_ver: str):
        return os.path.join(self.backup_dir, '{}{}'.format(
            self.backup_name_prefix, src_ver))

    def _backup_name_ext(self, src_ver: str):
        return '{}.{}'.format(self._backup_name(src_ver), self.backup_format)

    def _create_backup(self, src_ver: str):
        logger.debug('Creating backup for {}'.format(src_ver))
        shutil.make_archive(self._backup_name(src_ver),
                            self.backup_format, self.backup_target)

    def _restore_from_backup(self, src_ver: str):
        logger.info('Restoring from backup for {}'.format(src_ver))
        for file_path in self.files_to_preserve:
            try:
                shutil.copy2(os.path.join(self.backup_target, file_path),
                             os.path.join(self.tmp_dir, file_path))
            except IOError as e:
                logger.warning('Copying {} failed due to {}'
                               .format(file_path, e))
        shutil.unpack_archive(self._backup_name_ext(
            src_ver), self.backup_target, self.backup_format)
        for file_path in self.files_to_preserve:
            try:
                shutil.copy2(os.path.join(self.tmp_dir, file_path),
                             os.path.join(self.backup_target, file_path))
            except IOError as e:
                logger.warning('Copying {} failed due to {}'
                               .format(file_path, e))
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def _get_backups(self):
        files = [os.path.join(self.backup_dir, bk_file)
                 for bk_file in os.listdir(self.backup_dir)]
        files = [bk_file for bk_file in files if os.path.isfile(
            bk_file) and self.backup_name_prefix in bk_file]
        return sorted(files, key=os.path.getmtime, reverse=True)

    def _remove_old_backups(self):
        logger.display('Removing old backups')
        files = self._get_backups()
        if len(files):
            files = files[self.backup_num:]
        for file in files:
            os.remove(file)

    def _do_migration(self, curr_src_ver: str, new_src_ver: str):
        migrate(curr_src_ver, new_src_ver, self.timeout)

    def _migrate(self, curr_src_ver: str, new_src_ver: str):
        try:
            self._create_backup(curr_src_ver)
            self._do_migration(curr_src_ver, new_src_ver)
        except Exception as e:
            self._restore_from_backup(curr_src_ver)
            raise e
        self._remove_old_backups()

    def _do_upgrade(
            self,
            pkg_name: str,
            curr_pkg_ver: PackageVersion,
            new_pkg_ver: PackageVersion,
            rollback: bool
    ):
        from indy_node.server.upgrader import Upgrader
        cur_node_src_ver = Upgrader.get_src_version()
        logger.info(
            "Trying to upgrade from {}={} to {}"
            .format(pkg_name, curr_pkg_ver, new_pkg_ver)
        )

        try:
            self._call_upgrade_script(pkg_name, new_pkg_ver)
            if migrate:
                # TODO test for migration, should fail if nocache is False !!!
                new_node_src_ver = Upgrader.get_src_version(nocache=True)
                # assumption: migrations are anchored to release versions only
                self._migrate(cur_node_src_ver.release, new_node_src_ver.release)
            self._call_restart_node_script()
        except Exception as ex:
            logger.error(
                "Upgrade from {} to {} failed: {}"
                .format(curr_pkg_ver, new_pkg_ver, ex)
            )
            if rollback:
                logger.error("Trying to rollback to the previous version {} {}"
                             .format(ex, curr_pkg_ver))
                self._do_upgrade(
                    pkg_name, new_pkg_ver, curr_pkg_ver, rollback=False)

    # TODO test
    def _upgrade(
            self,
            new_src_ver: SourceVersion,
            pkg_name: str,
            migrate=True,
            rollback=True
    ):
        curr_pkg_ver, _ = NodeControlUtil.curr_pkg_info(pkg_name)

        if not curr_pkg_ver:
            logger.error("package {} is not found".format(pkg_name))
            return

        logger.info(
            "Current version of package '{}' is '{}'"
            .format(pkg_name, curr_pkg_ver)
        )

        new_pkg_ver = NodeControlUtil.get_latest_pkg_version(
            pkg_name, upstream=new_src_ver)
        if not new_pkg_ver:
            logger.error(
                "Upgrade from {} to upstream version {} failed: package {} for"
                " upstream version is not found"
                .format(curr_pkg_ver, new_src_ver, pkg_name)
            )
            return

        self._do_upgrade(pkg_name, curr_pkg_ver, new_pkg_ver, rollback)

    def _restart(self):
        try:
            self._call_restart_node_script()
        except Exception as ex:
            logger.error("Restart fail: " + ex.args[0])

    def _process_data(self, data):
        import json
        try:
            command = json.loads(data.decode("utf-8"))
            logger.debug("Decoded ", command)
            if command[MESSAGE_TYPE] == UPGRADE_MESSAGE:
                pkg_name = command['pkg_name']
                upstream_cls = src_version_cls(pkg_name)
                try:
                    new_src_ver = upstream_cls(command['version'])
                except InvalidVersionError as exc:
                    logger.error(
                        "invalid version {} for package {} with upstream class {}: {}"
                        .format(command['version'], pkg_name, upstream_cls, exc)
                    )
                else:
                    self._upgrade(new_src_ver, pkg_name)
            elif command[MESSAGE_TYPE] == RESTART_MESSAGE:
                self._restart()
        except json.decoder.JSONDecodeError as e:
            logger.error("JSON decoding failed: {}".format(e))
        except Exception as e:
            logger.error("Unexpected error in _process_data {}".format(e))

    def start(self):
        NodeControlUtil.hold_packages(self.config.PACKAGES_TO_HOLD + self.hold_ext)

        # Sockets from which we expect to read
        readers = [self.server]

        # Sockets to which we expect to write
        writers = []
        errs = []

        while readers:
            # Wait for at least one of the sockets to be ready for processing
            logger.debug('Waiting for the next event')
            readable, writable, exceptional = select.select(
                readers, writers, errs)
            for s in readable:
                if s is self.server:
                    # A "readable" server socket is ready to accept a
                    # connection
                    connection, client_address = s.accept()
                    logger.debug('New connection from {} on fd {}'
                                 .format(client_address, connection.fileno()))
                    connection.setblocking(0)
                    readers.append(connection)
                else:
                    data = s.recv(8192)
                    if data:
                        logger.debug(
                            'Received "{}" from {} on fd {}'
                            .format(data, s.getpeername(), s.fileno()))
                        self._process_data(data)
                    else:
                        logger.debug('Closing socket with fd {}'
                                     .format(s.fileno()))
                        readers.remove(s)
                        s.close()



from indy_common.config_helper import NodeConfigHelper
from indy_node.test.conftest import sdk_node_theta_added
from indy_node.test.helper import TestNode
from plenum.test.test_node import ensureElectionsDone

from plenum.common.util import randomString, hexToFriendly
from plenum.common.constants import SERVICES, TARGET_NYM, DATA
from plenum.common.txn_util import get_payload_data
from plenum.test.helper import waitForViewChange

from plenum.test.pool_transactions.helper import demote_node, promote_node

nodeCount = 7


def test_suspend_node(looper, sdk_pool_handle, sdk_wallet_trustee, nodeSet,
                    tdir, tconf, allPluginsPath):
    """
    Suspend a node and then cancel suspension. Suspend while suspended
    to test that there is no error
    """
    start_view_no = nodeSet[0].viewNo
    new_steward_wallet, new_node = sdk_node_theta_added(looper,
                                                        nodeSet,
                                                        tdir,
                                                        tconf,
                                                        sdk_pool_handle,
                                                        sdk_wallet_trustee,
                                                        allPluginsPath,
                                                        node_config_helper_class=NodeConfigHelper,
                                                        testNodeClass=TestNode,
                                                        name="Node-" + randomString(5))
    waitForViewChange(looper=looper, txnPoolNodeSet=nodeSet,
                      expectedViewNo=start_view_no + 1)
    ensureElectionsDone(looper=looper, nodes=nodeSet)

    demote_node(looper, sdk_wallet_trustee, sdk_pool_handle, new_node)
    _wait_view_change_finish(looper, nodeSet[:-1], start_view_no + 1)

    demote_node(looper, sdk_wallet_trustee, sdk_pool_handle, new_node)

    promote_node(looper, sdk_wallet_trustee, sdk_pool_handle, new_node)
    _wait_view_change_finish(looper, nodeSet[:-1], start_view_no + 2)

    promote_node(looper, sdk_wallet_trustee, sdk_pool_handle, new_node)


def _wait_view_change_finish(looper, nodes, view_no):
    waitForViewChange(looper=looper, txnPoolNodeSet=nodes,
                      expectedViewNo=view_no)
    ensureElectionsDone(looper=looper, nodes=nodes)


def testDemoteNodeWhichWasNeverActive(looper, nodeSet, sdk_pool_handle,
                                      sdk_wallet_trustee, tdir, tconf,
                                      allPluginsPath):
    """
    Add a node without services field and check that the ledger does not
    contain the `services` field and check that it can be demoted and
    the ledger has `services` as empty list
    """
    new_steward_wallet, new_node = sdk_node_theta_added(looper,
                                                        nodeSet,
                                                        tdir,
                                                        tconf,
                                                        sdk_pool_handle,
                                                        sdk_wallet_trustee,
                                                        allPluginsPath,
                                                        node_config_helper_class=NodeConfigHelper,
                                                        testNodeClass=TestNode,
                                                        name="Node-" + randomString(5),
                                                        services=None)

    looper.runFor(tconf.PROPAGATE_REQUEST_DELAY * 1.5)
    for node in nodeSet[:nodeCount]:
        txn = [t for _, t in node.poolLedger.getAllTxn()][-1]
        txn_data = get_payload_data(txn)
        assert txn_data[TARGET_NYM] == hexToFriendly(new_node.nodestack.verhex)
        assert SERVICES not in txn_data[DATA]

    demote_node(looper, new_steward_wallet, sdk_pool_handle, new_node)

    for node in nodeSet[:nodeCount]:
        txn = [t for _, t in node.poolLedger.getAllTxn()][-1]
        txn_data = get_payload_data(txn)
        assert txn_data[TARGET_NYM] == hexToFriendly(new_node.nodestack.verhex)
        assert SERVICES in txn_data[DATA] and txn_data[DATA][SERVICES] == []



import dateutil.parser

from indy_common.constants import IN_PROGRESS
from indy_node.server.upgrade_log import UpgradeLogData

from indy_node.test.upgrade.helper import check_node_sent_acknowledges_upgrade, clear_config_ledger

whitelist = ['unable to send message']


def test_node_sent_upgrade_in_progress(looper, nodeSet, nodeIds, validUpgrade):
    '''
    Test that each node sends NODE_UPGRADE In Progress event
    (because it sees scheduledUpgrade in the Upgrader)
    '''
    clear_config_ledger(nodeSet)
    version = validUpgrade['version']
    for node in nodeSet:
        node_id = node.poolManager.get_nym_by_name(node.name)
        node.upgrader.scheduledAction = UpgradeLogData(
            dateutil.parser.parse(validUpgrade['schedule'][node_id]),
            version,
            "upgrade_id",
            validUpgrade['package']
        )
        node.notify_upgrade_start()

    check_node_sent_acknowledges_upgrade(looper, nodeSet, nodeIds,
                                         allowed_actions=[IN_PROGRESS],
                                         ledger_size=len(nodeSet),
                                         expected_version=version)



from collections import OrderedDict
from indy_common.types import RevocRegEntryValueField, ClientRevocRegEntrySubmitField
from plenum.common.messages.fields import NonEmptyStringField, IterableField, ConstantField, LimitedLengthStringField
from indy_common.constants import REVOC_REG_DEF_ID, REVOC_TYPE, VALUE, PREV_ACCUM, \
    ACCUM, ISSUED, REVOKED, TXN_TYPE


EXPECTED_REVOC_DEF_ENTRY_VALUE_FIELDS = OrderedDict([
    (PREV_ACCUM, NonEmptyStringField),
    (ACCUM, NonEmptyStringField),
    (ISSUED, IterableField),
    (REVOKED, IterableField),
])

EXPECTED_REVOC_DEF_ENTRY_SUBMIT_FIELDS = OrderedDict([
    (TXN_TYPE, ConstantField),
    (REVOC_REG_DEF_ID, NonEmptyStringField),
    (REVOC_TYPE, LimitedLengthStringField),
    (VALUE, RevocRegEntryValueField)
])


def test_revoc_entry_value_has_expected_fields():
    actual_field_names = OrderedDict(
        RevocRegEntryValueField.schema).keys()
    assert actual_field_names == EXPECTED_REVOC_DEF_ENTRY_VALUE_FIELDS.keys()


def test_revoc_entry_value_has_expected_validators():
    schema = dict(RevocRegEntryValueField.schema)
    for field, validator in EXPECTED_REVOC_DEF_ENTRY_VALUE_FIELDS.items():
        assert isinstance(schema[field], validator)


def test_client_revoc_entry_submit_has_expected_fields():
    actual_field_names = OrderedDict(
        ClientRevocRegEntrySubmitField.schema).keys()
    assert actual_field_names == EXPECTED_REVOC_DEF_ENTRY_SUBMIT_FIELDS.keys()


def test_client_revoc_entry_submit_has_expected_validators():
    schema = dict(ClientRevocRegEntrySubmitField.schema)
    for field, validator in EXPECTED_REVOC_DEF_ENTRY_SUBMIT_FIELDS.items():
        assert isinstance(schema[field], validator)



import functools

import pytest
import time
import json

from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.state import config
from plenum.common.constants import NYM, ROLE, TRUSTEE, STEWARD, CONFIG_LEDGER_ID, TXN_METADATA, \
    TXN_METADATA_SEQ_NO
from plenum.common.txn_util import reqToTxn, get_payload_data
from plenum.common.request import Request
from plenum.server.ledgers_bootstrap import LedgersBootstrap
from plenum.test.testing_utils import FakeSomething
from state.pruning_state import PruningState
from storage.kv_in_memory import KeyValueStorageInMemory

from indy_node.test.helper import build_auth_rule_request_json


def reset_state(db_manager, lid):
    db_manager.databases[lid].state = PruningState(KeyValueStorageInMemory())
    db_manager._init_db_list()
    return db_manager


@pytest.fixture
def prepare_request(looper, sdk_wallet_trustee):
    constraint = AuthConstraint(TRUSTEE,
                                1)
    action = AuthActionAdd(txn_type=NYM,
                           field=ROLE,
                           value=STEWARD)
    auth_params = {
        "constraint": constraint.as_dict,
        "auth_action": "ADD",
        "auth_type": NYM,
        'field': ROLE,
        "new_value": STEWARD
    }
    req_json = build_auth_rule_request_json(
        looper, sdk_wallet_trustee[1], **auth_params
    )
    return action, constraint, Request(**json.loads(req_json))


def test_update_state(write_manager,
                      db_manager,
                      prepare_request,
                      constraint_serializer):
    action, constraint, request = prepare_request
    head_hash_before = db_manager.get_state(CONFIG_LEDGER_ID).headHash
    txn = reqToTxn(request)
    write_manager.update_state(txn, isCommitted=False)
    head_hash_after = db_manager.get_state(CONFIG_LEDGER_ID).headHash
    from_state = db_manager.get_state(CONFIG_LEDGER_ID).get(
        config.make_state_path_for_auth_rule(action.get_action_id()),
        isCommitted=False)
    assert constraint_serializer.deserialize(from_state) == constraint
    assert head_hash_after != head_hash_before


def test_uncommitted_state_after_apply(write_manager,
                                       db_manager,
                                       constraint_serializer,
                                       prepare_request):
    action, constraint, request = prepare_request

    write_manager.apply_request(request, int(time.time()))
    assert len(db_manager.get_ledger(CONFIG_LEDGER_ID).uncommittedTxns) == 1
    assert db_manager.get_state(CONFIG_LEDGER_ID).committedHeadHash != \
        db_manager.get_state(CONFIG_LEDGER_ID).headHash
    from_state = db_manager.get_state(CONFIG_LEDGER_ID).get(
        config.make_state_path_for_auth_rule(action.get_action_id()),
        isCommitted=False)
    assert constraint_serializer.deserialize(from_state) == constraint


def test_init_state_from_ledger(write_manager,
                                db_manager,
                                constraint_serializer,
                                prepare_request):
    reset_state(db_manager, CONFIG_LEDGER_ID)
    req_count = 1
    action, constraint, request = prepare_request
    txn = reqToTxn(request)
    txn[TXN_METADATA][TXN_METADATA_SEQ_NO] = 1
    """Add txn to ledger"""
    db_manager.get_ledger(CONFIG_LEDGER_ID).appendTxns([txn])
    db_manager.get_ledger(CONFIG_LEDGER_ID).commitTxns(req_count)
    # ToDo: ugly fix... Refactor this on pluggable req handler integration phase
    init_state_from_ledger = functools.partial(
        LedgersBootstrap._init_state_from_ledger,
        FakeSomething(
            db_manager=db_manager,
            write_manager=write_manager,
            _update_txn_with_extra_data=lambda txn: txn))
    """Check that txn is not exist in state"""
    assert db_manager.get_state(CONFIG_LEDGER_ID).get(config.make_state_path_for_auth_rule(action.get_action_id()),
                            isCommitted=False) is None
    assert db_manager.get_state(CONFIG_LEDGER_ID).get(config.make_state_path_for_auth_rule(action.get_action_id()),
                            isCommitted=True) is None
    txns_from_ledger = [t for t in db_manager.get_ledger(CONFIG_LEDGER_ID).getAllTxn()]
    """Check, that txn exist in ledger"""
    assert len(txns_from_ledger) == 1
    assert get_payload_data(txns_from_ledger[0][1]) == get_payload_data(txn)
    """Emulating node starting"""
    init_state_from_ledger(CONFIG_LEDGER_ID)
    """Check that txn was added into state"""
    from_state = db_manager.get_state(CONFIG_LEDGER_ID) .get(
        config.make_state_path_for_auth_rule(action.get_action_id()),
        isCommitted=True)
    assert constraint_serializer.deserialize(from_state) == constraint



#!/usr/bin/python3.5
import os
import shutil
import pwd
import grp
import stat

from stp_core.common.log import getlogger
from indy_common.config_util import getConfigOnce
from indy_common.config_helper import ConfigHelper

import indy_node.general_config.indy_config as indy_config


logger = getlogger()


def ext_copytree(src, dst):
    if not os.path.isdir(dst):
        shutil.copytree(src, dst)


def set_own_perm(usr, dir_list):
    uid = pwd.getpwnam(usr).pw_uid
    gid = grp.getgrnam(usr).gr_gid
    perm_mask_rw = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP
    perm_mask_rwx = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP

    for cdir in dir_list:
        os.chown(cdir, uid, gid)
        os.chmod(cdir, perm_mask_rwx)
        for croot, sub_dirs, cfiles in os.walk(cdir):
            for fs_name in sub_dirs:
                os.chown(os.path.join(croot, fs_name), uid, gid)
                os.chmod(os.path.join(croot, fs_name), perm_mask_rwx)
            for fs_name in cfiles:
                os.chown(os.path.join(croot, fs_name), uid, gid)
                os.chmod(os.path.join(croot, fs_name), perm_mask_rw)


def migrate_nodes_data():
    config = getConfigOnce()
    config_helper = ConfigHelper(config)

    # Move data
    old_nodes_data_dir = os.path.join(config_helper.ledger_data_dir, 'nodes')
    new_node_data_dir = config_helper.ledger_data_dir
    try:
        visit_dirs = os.listdir(old_nodes_data_dir)
    except FileNotFoundError:
        visit_dirs = []
    for node_name in visit_dirs:
        move_path = os.path.join(old_nodes_data_dir, node_name)
        to_path = os.path.join(new_node_data_dir, node_name)
        ext_copytree(move_path, to_path)
    shutil.rmtree(old_nodes_data_dir)
    set_own_perm("indy", [new_node_data_dir])


def migrate_config():
    config = getConfigOnce()
    general_config_path = os.path.join(config.GENERAL_CONFIG_DIR, config.GENERAL_CONFIG_FILE)

    if not os.path.exists(general_config_path):
        return

    if not os.path.isfile(general_config_path):
        logger.error("General config '{}' exists, but it is not a regular file, abort migration."
                     .format(general_config_path))
    else:
        logger.info("Open '{}' for appending missing paths configuration."
                    .format(general_config_path))
        general_config_file = open(general_config_path, "a")
        logger.info("Open '{}' to get missing paths configuration."
                    .format(indy_config.__file__))
        indy_config_file = open(indy_config.__file__, "r")

        general_config_file.write("\n")

        for line in indy_config_file:
            if not line.startswith("NETWORK_NAME") and not line == "# Current network\n":
                logger.info("Append '{}' line to '{}'."
                            .format(line, general_config_path))
                general_config_file.write(line)

        general_config_file.close()
        indy_config_file.close()


def migrate_all():
    # Migrate config first!
    migrate_config()
    migrate_nodes_data()


migrate_all()



import pytest
from indy_common.authorize.auth_constraints import OFF_LEDGER_SIGNATURE
from indy_node.server.request_handlers.config_req_handlers.auth_rule.abstract_auth_rule_handler import \
    AbstractAuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rule_handler import AuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rule_handler_1_9_1 import AuthRuleHandler191
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from plenum.common.txn_util import reqToTxn, get_payload_data


@pytest.fixture(scope="module")
def auth_rule_handler_1_9_1(db_manager, write_auth_req_validator):
    return AuthRuleHandler191(db_manager, write_auth_req_validator)


def test_update_state(auth_rule_request, auth_rule_handler_1_9_1: AuthRuleHandler):
    txn = reqToTxn(auth_rule_request)
    payload = get_payload_data(txn)
    constraint = StaticAuthRuleHelper.get_auth_constraint(payload)
    auth_key = StaticAuthRuleHelper.get_auth_key(payload)
    path = AbstractAuthRuleHandler.make_state_path_for_auth_rule(auth_key)

    auth_rule_handler_1_9_1.update_state(txn, None, auth_rule_request)
    expected_dict = constraint.as_dict
    expected_dict[OFF_LEDGER_SIGNATURE] = False
    assert auth_rule_handler_1_9_1.get_from_state(path) == expected_dict




from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.util import randomString
from plenum.test.node_catchup.helper import checkNodeDataForEquality, \
    waitNodeDataEquality
from plenum.test.pool_transactions.helper import sdk_add_new_nym, \
    disconnect_node_and_ensure_disconnected
from plenum.test.test_node import ensure_node_disconnected, checkNodesConnected
from indy_node.test.conftest import sdk_node_theta_added
from indy_node.test.helper import TestNode, sdk_add_raw_attribute, \
    start_stopped_node
from indy_common.config_helper import NodeConfigHelper


def test_new_node_catchup_update_projection(looper,
                                            nodeSet, tconf, tdir,
                                            sdk_pool_handle,
                                            sdk_wallet_trustee,
                                            allPluginsPath,
                                            some_transactions_done):
    """
    A node which receives txns from catchup updates both ledger and projection
    4 nodes start up and some txns happen, after txns are done, new node joins
    and starts catching up, the node should not process requests while catchup
    is in progress. Make sure the new requests are coming from the new NYMs
    added while the node was offline or catching up.
    """
    # Create a new node and stop it.

    new_steward_wallet, new_node = sdk_node_theta_added(looper,
                                                        nodeSet,
                                                        tdir,
                                                        tconf,
                                                        sdk_pool_handle,
                                                        sdk_wallet_trustee,
                                                        allPluginsPath,
                                                        node_config_helper_class=NodeConfigHelper,
                                                        testNodeClass=TestNode)

    waitNodeDataEquality(looper, new_node, *nodeSet[:-1])
    ta_count = 2
    np_count = 2
    new_txn_count = 2 * ta_count + np_count  # Since ATTRIB txn is done for TA
    old_ledger_sizes = {}
    new_ledger_sizes = {}
    old_projection_sizes = {}
    new_projection_sizes = {}
    old_seq_no_map_sizes = {}
    new_seq_no_map_sizes = {}

    def get_ledger_size(node):
        return len(node.domainLedger)

    def get_projection_size(node):
        domain_state = node.getState(DOMAIN_LEDGER_ID)
        return len(domain_state.as_dict)

    def get_seq_no_map_size(node):
        return node.seqNoDB.size

    def fill_counters(ls, ps, ss, nodes):
        for n in nodes:
            ls[n.name] = get_ledger_size(n)
            ps[n.name] = get_projection_size(n)
            ss[n.name] = get_seq_no_map_size(n)

    def check_sizes(nodes):
        for node in nodes:
            assert new_ledger_sizes[node.name] - \
                   old_ledger_sizes[node.name] == new_txn_count
            assert new_projection_sizes[node.name] - \
                   old_projection_sizes[node.name] == new_txn_count
            assert new_seq_no_map_sizes[node.name] - \
                   old_seq_no_map_sizes[node.name] == new_txn_count * 2

    # Stop a node and note down the sizes of ledger and projection (state)
    other_nodes = nodeSet[:-1]
    fill_counters(old_ledger_sizes, old_projection_sizes, old_seq_no_map_sizes,
                  other_nodes)
    new_node.cleanupOnStopping = False
    # new_node.stop()
    # looper.removeProdable(new_node)
    # ensure_node_disconnected(looper, new_node, other_nodes)

    disconnect_node_and_ensure_disconnected(looper,
                                            nodeSet,
                                            new_node.name)
    looper.removeProdable(name=new_node.name)
    endorsers = []
    attributes = []
    for i in range(ta_count):
        endorsers.append(
            sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                            role='ENDORSER', alias='TA' + str(i)))
        attributes.append((randomString(6), randomString(10)))
        sdk_add_raw_attribute(looper, sdk_pool_handle, endorsers[-1],
                              *attributes[-1])
    non_privileged = []
    for i in range(np_count):
        non_privileged.append(
            sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                            alias='NP' + str(i)))

    checkNodeDataForEquality(nodeSet[0], *other_nodes)
    fill_counters(new_ledger_sizes, new_projection_sizes, new_seq_no_map_sizes,
                  other_nodes)
    # The size difference should be same as number of new NYM txns
    check_sizes(other_nodes)

    new_node = start_stopped_node(new_node, looper, tconf,
                                  tdir, allPluginsPath)
    nodeSet[-1] = new_node

    fill_counters(old_ledger_sizes, old_projection_sizes, old_seq_no_map_sizes,
                  [new_node])
    looper.run(checkNodesConnected(nodeSet))
    waitNodeDataEquality(looper, new_node, *other_nodes)
    fill_counters(new_ledger_sizes, new_projection_sizes, new_seq_no_map_sizes,
                  [new_node])
    check_sizes([new_node])

    # Set the old counters to be current ledger and projection size
    fill_counters(old_ledger_sizes, old_projection_sizes, old_seq_no_map_sizes,
                  nodeSet)

    more_nyms_count = 2
    for wh in endorsers:
        for i in range(more_nyms_count):
            non_privileged.append(sdk_add_new_nym(looper, sdk_pool_handle, wh,
                                                  alias='NP1' + str(i)))

    # The new node should process transactions done by Nyms added to its
    # ledger while catchup
    fill_counters(new_ledger_sizes, new_projection_sizes, new_seq_no_map_sizes,
                  nodeSet)
    new_txn_count = more_nyms_count * len(endorsers)
    check_sizes(nodeSet)



from copy import deepcopy

from indy_common.constants import ACTION, CANCEL, SCHEDULE
from indy_node.test import waits
from indy_node.test.upgrade.helper import checkNoUpgradeScheduled, \
    sdk_ensure_upgrade_sent
from plenum.test.test_node import checkNodesConnected, ensureElectionsDone
from stp_core.loop.eventually import eventually
from indy_common.config_helper import NodeConfigHelper


def test_node_does_not_reschedule_cancelled_upgrade_after_restart(
        upgradeScheduled, looper, nodeSet, validUpgrade,
        testNodeClass, tdir, tconf, allPluginsPath,
        sdk_pool_handle, sdk_wallet_trustee):
    # Cancel the scheduled upgrade
    valid_upgrade_cancel = deepcopy(validUpgrade)
    valid_upgrade_cancel[ACTION] = CANCEL
    del valid_upgrade_cancel[SCHEDULE]

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, valid_upgrade_cancel)

    # Verify that no upgrade is scheduled now
    looper.run(
        eventually(
            checkNoUpgradeScheduled,
            nodeSet,
            retryWait=1,
            timeout=waits.expectedNoUpgradeScheduled()))

    # Restart all the nodes
    names = []
    while nodeSet:
        node = nodeSet.pop()
        names.append(node.name)
        node.cleanupOnStopping = False
        looper.removeProdable(node)
        node.stop()
        del node

    for nm in names:
        config_helper = NodeConfigHelper(nm, tconf, chroot=tdir)
        node = testNodeClass(nm, config_helper=config_helper,
                             config=tconf, pluginPaths=allPluginsPath)
        looper.add(node)
        nodeSet.append(node)

    looper.run(checkNodesConnected(nodeSet))
    ensureElectionsDone(looper=looper, nodes=nodeSet, retryWait=1)

    # Verify that still no upgrade is scheduled
    looper.run(
        eventually(
            checkNoUpgradeScheduled,
            nodeSet,
            retryWait=1,
            timeout=waits.expectedNoUpgradeScheduled()))



import datetime
import os
import random
from typing import Tuple, Union, TypeVar, List, Callable

import libnacl.secret
from base58 import b58decode
from common.serializers.serialization import serialize_msg_for_signing
from plenum.common.types import f
from plenum.common.util import isHex, cryptonymToHex
from common.error import error
from stp_core.crypto.nacl_wrappers import Verifier


def getMsgWithoutSig(msg, sigFieldName=f.SIG.nm):
    msgWithoutSig = {}
    for k, v in msg.items():
        if k != sigFieldName:
            msgWithoutSig[k] = v
    return msgWithoutSig


def verifySig(identifier, signature, msg) -> bool:
    key = cryptonymToHex(identifier) if not isHex(
        identifier) else identifier
    ser = serialize_msg_for_signing(msg)
    b64sig = signature.encode('utf-8')
    sig = b58decode(b64sig)
    vr = Verifier(key)
    return vr.verify(sig, ser)


def getSymmetricallyEncryptedVal(val, secretKey: Union[str, bytes] = None) -> \
        Tuple[str, str]:
    """
    Encrypt the provided value with symmetric encryption

    :param val: the value to encrypt
    :param secretKey: Optional key, if provided should be either in hex or bytes
    :return: Tuple of the encrypted value and secret key encoded in hex
    """

    if isinstance(val, str):
        val = val.encode("utf-8")
    if secretKey:
        if isHex(secretKey):
            secretKey = bytes(bytearray.fromhex(secretKey))
        elif not isinstance(secretKey, bytes):
            error("Secret key must be either in hex or bytes")
        box = libnacl.secret.SecretBox(secretKey)
    else:
        box = libnacl.secret.SecretBox()

    return box.encrypt(val).hex(), box.sk.hex()


def getSymmetricallyDecryptedVal(val, secretKey: Union[str, bytes]) -> str:
    if isHex(val):
        val = bytes(bytearray.fromhex(val))
    elif isinstance(val, str):
        val = val.encode("utf-8")
    if isHex(secretKey):
        secretKey = bytes(bytearray.fromhex(secretKey))
    elif isinstance(secretKey, str):
        secretKey = secretKey.encode()
    box = libnacl.secret.SecretBox(secretKey)
    return box.decrypt(val).decode()


def dateTimeEncoding(obj):
    if isinstance(obj, datetime.datetime):
        return int(obj.strftime('%s'))
    raise TypeError('Not sure how to serialize %s' % (obj,))


def getNonce(length=32):
    hexChars = [hex(i)[2:] for i in range(0, 16)]
    return "".join([random.choice(hexChars) for i in range(length)])


def get_reply_if_confirmed(client, identifier, request_id: int):
    reply, status = client.getReply(identifier, request_id)
    if status == 'CONFIRMED':
        return reply, None
    _, errors = \
        client.reqRepStore.getAllReplies(identifier, request_id)
    if not errors:
        return None, None
    sender, error_reason = errors.popitem()
    return reply, error_reason


# TODO: Should have a timeout, should not have kwargs
def ensureReqCompleted(
        loop,
        reqKey,
        client,
        clbk=None,
        pargs=None,
        kwargs=None,
        cond=None):

    reply, err = get_reply_if_confirmed(client, *reqKey)

    if err is None and reply is None and (cond is None or not cond()):
        loop.call_later(.2, ensureReqCompleted, loop,
                        reqKey, client, clbk, pargs, kwargs, cond)
    elif clbk:
        # TODO: Do something which makes reply and error optional in the
        # callback.
        # TODO: This is kludgy, but will be resolved once we move away from
        # this callback pattern
        if pargs is not None and kwargs is not None:
            clbk(reply, err, *pargs, **kwargs)
        elif pargs is not None and kwargs is None:
            clbk(reply, err, *pargs)
        elif pargs is None and kwargs is not None:
            clbk(reply, err, **kwargs)
        else:
            clbk(reply, err)


def getNonceForProof(nonce):
    return int(nonce, 16)


T = TypeVar('T')


def getIndex(predicateFn: Callable[[T], bool], items: List[T]) -> int:
    """
    Finds the index of an item in list, which satisfies predicate
    :param predicateFn: predicate function to run on items of list
    :param items: list of tuples
    :return: first index for which predicate function returns True
    """
    try:
        return next(i for i, v in enumerate(items) if predicateFn(v))
    except StopIteration:
        return -1


def compose_cmd(cmd):
    if os.name != 'nt':
        cmd = ' '.join(cmd)
    return cmd


def invalidate_config_caches():
    import stp_core.common.config.util
    import plenum.common.config_util
    import indy_common.config_util

    # All 3 references must be nullified because all they reference
    # the same object due to specific logic of getConfig methods
    stp_core.common.config.util.CONFIG = None
    plenum.common.config_util.CONFIG = None
    indy_common.config_util.CONFIG = None



import json

import pytest
from indy.ledger import build_acceptance_mechanisms_request

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.types import Request
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.helper import build_auth_rule_request_json
from plenum.common.constants import TXN_AUTHOR_AGREEMENT_DISABLE, TXN_TYPE
from plenum.common.exceptions import RequestRejectedException
from plenum.common.util import randomString, get_utc_epoch
from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_request_from_dict
from plenum.test.pool_transactions.helper import sdk_add_new_nym, sdk_sign_and_send_prepared_request
from plenum.test.txn_author_agreement.helper import sdk_send_txn_author_agreement


class TAADisableTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        req = self.taa_aml_request()
        rep = sdk_sign_and_send_prepared_request(self.looper, self.trustee_wallet, self.sdk_pool_handle, req)
        sdk_get_and_check_replies(self.looper, [rep])
        self.send_taa()

    def taa_aml_request(self):
        return self.looper.loop.run_until_complete(build_acceptance_mechanisms_request(
            self.trustee_wallet[1],
            json.dumps({
                'Nice way': 'very good way to accept agreement'}),
            randomString(), randomString()))

    def send_taa(self):
        sdk_send_txn_author_agreement(self.looper, self.sdk_pool_handle, self.trustee_wallet,
                                      randomString(10), randomString(5), ratified=get_utc_epoch())

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet,
                                                  role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=TXN_AUTHOR_AGREEMENT_DISABLE,
            field='*',
            new_value='*',
            constraint=constraint.as_dict
        )

    def send_taa_disable_req(self, wallet):
        operation = {TXN_TYPE: TXN_AUTHOR_AGREEMENT_DISABLE}
        req = sdk_sign_request_from_dict(self.looper, wallet, operation)
        self.send_and_check(json.dumps(req), wallet)

    def run(self):
        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            self.send_taa_disable_req(self.trustee_wallet)

        # Step 3. Check, that new auth rule is used
        self.send_taa_disable_req(self.new_default_wallet)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.send_taa()
        self.send_taa_disable_req(self.trustee_wallet)

    def result(self):
        pass

    def down(self):
        pass



import pytest

from indy_common.constants import APP_NAME
from indy_node.utils.node_control_tool import NodeControlTool
from plenum.test.helper import randomText
from indy_node.utils.node_control_utils import NodeControlUtil

EXT_PKT_VERSION = '7.88.999'
EXT_PKT_NAME = 'SomeTopLevelPkt'
node_package = (APP_NAME, '0.0.1')
EXT_TOP_PKT_DEPS = [("aa", "1.1.1"), ("bb", "2.2.2")]
PACKAGE_MNG_EXT_PTK_OUTPUT = "Package: {}\nStatus: install ok installed\nPriority: extra\nSection: default\n" \
                             "Installed-Size: 21\nMaintainer: EXT_PKT_NAME-fond\nArchitecture: amd64\nVersion: {}\n" \
                             "Depends: {}, {} (= {}), {} (= {})\nDescription: EXT_PKT_DEPS-desc\n" \
                             "License: EXT_PKT_DEPS-lic\nVendor: none\n". \
    format(EXT_PKT_NAME, EXT_PKT_VERSION, APP_NAME, *EXT_TOP_PKT_DEPS[0], *EXT_TOP_PKT_DEPS[1])


@pytest.fixture(scope="module")
def tconf(tconf, tdir):
    oldv = tconf.UPGRADE_ENTRY
    tconf.UPGRADE_ENTRY = EXT_PKT_NAME
    tconf.LOG_DIR = tdir
    yield tconf
    tconf.UPGRADE_ENTRY = oldv


def test_node_as_depend(monkeypatch, tconf):
    nct = NodeControlTool(config=tconf)
    top_level_package = (EXT_PKT_NAME, EXT_PKT_VERSION)
    plenum_package = ('indy-plenum', '0.0.3')
    top_level_package_with_version = '{}={}'.format(*top_level_package)
    top_level_package_dep1_with_version = '{}={}'.format(*EXT_TOP_PKT_DEPS[0])
    top_level_package_dep2_with_version = '{}={}'.format(*EXT_TOP_PKT_DEPS[1])
    node_package_with_version = '{}={}'.format(*node_package)
    plenum_package_with_version = '{}={}'.format(*plenum_package)
    mock_info = {
        top_level_package_with_version: "{}\nVersion:{}\nDepends:{} (= {}), {} (= {}), {} (= {})\n".format(
            randomText(100), top_level_package[1], *node_package, *EXT_TOP_PKT_DEPS[0], *EXT_TOP_PKT_DEPS[1]),
        node_package_with_version: '{}\nVersion:{}\nDepends:{} (= {})\n'.format(
            randomText(100), node_package[1], *plenum_package),
        plenum_package_with_version: '{}\nVersion:{}\nDepends:{} (= {})\n'.format(
            randomText(100), plenum_package[1], *plenum_package),
        top_level_package_dep1_with_version: '{}\nVersion:{}\nDepends:{} (= {})\n'.format(
            randomText(100), EXT_TOP_PKT_DEPS[0][1], *plenum_package),
        top_level_package_dep2_with_version: '{}\nVersion:{}\nDepends:{} (= {})\n'.format(
            randomText(100), EXT_TOP_PKT_DEPS[1][1], *node_package),
    }

    def mock_get_info_from_package_manager(*package):
        ret = ""
        for p in package:
            ret += mock_info.get(p, "")
        return ret

    monkeypatch.setattr(NodeControlUtil, 'update_package_cache', lambda *x: None)
    monkeypatch.setattr(NodeControlUtil, '_get_info_from_package_manager',
                        lambda *x: mock_get_info_from_package_manager(*x))
    monkeypatch.setattr(NodeControlUtil, 'get_sys_holds',
                        lambda *x: [top_level_package[0], plenum_package[0], node_package[0],
                                    EXT_TOP_PKT_DEPS[0][0], EXT_TOP_PKT_DEPS[1][0]])
    monkeypatch.setattr(NodeControlUtil, '_get_curr_info', lambda *x: PACKAGE_MNG_EXT_PTK_OUTPUT)
    ret = nct._get_deps_list(top_level_package_with_version)
    nct.server.close()
    assert sorted(ret.split()) == sorted([plenum_package_with_version,
                                          node_package_with_version, top_level_package_dep2_with_version,
                                          top_level_package_dep1_with_version, top_level_package_with_version])



import subprocess
import shutil
import codecs
import locale
import re
from typing import Iterable, Type, Tuple, List, Union


from stp_core.common.log import getlogger
from common.version import (
    InvalidVersionError, SourceVersion, PackageVersion, GenericVersion
)

from indy_common.version import src_version_cls
from indy_common.util import compose_cmd


# Package manager command output could contain some utf-8 symbols
# to handle such a case automatic stream parsing is prohibited,
# decode error handler is added, proper decoder is selected

# copied from validator-info from plenum
def decode_err_handler(error):
    length = error.end - error.start
    return length * ' ', error.end


# copied from validator-info from plenum
codecs.register_error('decode_errors', decode_err_handler)


logger = getlogger()
TIMEOUT = 600
MAX_DEPS_DEPTH = 6


class ShellError(subprocess.CalledProcessError):
    def __init__(self, *args, exc: subprocess.CalledProcessError = None, **kwargs):
        if exc:
            super().__init__(exc.returncode, exc.cmd, output=exc.output, stderr=exc.stderr)
        else:
            super().__init__(*args, **kwargs)

    @property
    def stdout_decoded(self):
        return (self.stdout.decode(locale.getpreferredencoding(), 'decode_errors')
                if self.stdout else "")

    @property
    def stderr_decoded(self):
        return (self.stdout.decode(locale.getpreferredencoding(), 'decode_errors')
                if self.stderr else "")


# TODO use some library instead of dpkg fpr version routine
class DebianVersion(PackageVersion):
    cache = {}  # seems not actually necessary
    # https://www.debian.org/doc/debian-policy/ch-controlfields.html#version
    re_version = re.compile(r'(?:([0-9]):)?([0-9][a-zA-Z0-9.+\-~]*)')

    @classmethod
    def _cmp(cls, v1, v2):
        if v1 == v2:
            return 0
        else:
            cmd = compose_cmd(['dpkg', '--compare-versions', v1, 'gt', v2])
            try:
                NodeControlUtil.run_shell_script_extended(cmd)
            except ShellError as exc:
                if exc.stderr:
                    raise
                else:
                    return -1
            else:
                return 1

    @classmethod
    def cmp(cls, v1, v2):
        key = (v1.full, v2.full)
        if key not in cls.cache:
            cls.cache[key] = cls._cmp(*key)
            cls.cache[key[::-1]] = cls.cache[key] * (-1)

        return cls.cache[key]

    @classmethod
    def clear_cache(cls):
        cls.cache.clear()

    def __init__(
        self,
        version: str,
        keep_tilde: bool = False,
        upstream_cls: Type[Union[SourceVersion, None]] = GenericVersion
    ):
        parsed = self._parse(version, keep_tilde, upstream_cls)
        if not parsed[1]:
            raise InvalidVersionError(
                "{} is not a valid debian version".format(version)
            )

        self._version = version
        self._epoch, self._upstream, self._revision = parsed

    def _parse(
            self,
            version: str,
            keep_tilde: bool,
            upstream_cls: Type[SourceVersion],
    ):
        epoch = None
        upstream = None
        revision = None

        match = re.fullmatch(self.re_version, version)
        if match:
            epoch = match.group(1)
            upstream = match.group(2)
            if upstream:
                if not keep_tilde:
                    upstream = upstream.replace('~', '.')
                # TODO improve regex instead
                parts = upstream.split('-')
                if len(parts) > 1:
                    upstream = '-'.join(parts[:-1])
                    revision = parts[-1]
        return (
            epoch,
            upstream_cls(upstream) if upstream else None,
            revision
        )

    @property
    def full(self) -> str:
        return self._version

    @property
    def parts(self) -> Iterable:
        return (self.epoch, self.upstream, self.revision)

    @property
    def release(self) -> str:
        return self.full

    @property
    def release_parts(self) -> Iterable:
        return self.parts

    @property
    def epoch(self):
        return self._epoch

    @property
    def upstream(self):
        return self._upstream

    @property
    def revision(self):
        return self._revision


class NodeControlUtil:
    # Method is used in case we are interested in command output
    # errors are ignored
    # only critical errors are logged to journalctl
    @classmethod
    def run_shell_command(cls, command, timeout=TIMEOUT):
        try:
            ret = subprocess.run(command, shell=True, check=True, stdout=subprocess.PIPE, timeout=timeout)
            ret_bytes = ret.stdout
        except subprocess.CalledProcessError as ex:
            ret_bytes = ex.output
        except Exception as ex:
            raise Exception("command {} failed with {}".format(command, ex))
        ret_msg = ret_bytes.decode(locale.getpreferredencoding(), 'decode_errors').strip() if ret_bytes else ""
        return ret_msg

    # Method is used in case we are NOT interested in command output
    # everything: command, errors, output etc are logged to journalctl
    @classmethod
    def run_shell_script(cls, command, timeout=TIMEOUT):
        subprocess.run(command, shell=True, timeout=timeout, check=True)

    # TODO actually this should replace `run_shell_script` but it needs
    # deep testing and verification since it impacts upgrade routine a lot
    @classmethod
    def run_shell_script_extended(
            cls, command, stdout=False, stderr=False,
            timeout=TIMEOUT, check=True):
        try:
            res = subprocess.run(
                command, shell=True, timeout=timeout, check=check,
                stdout=None if stdout else subprocess.PIPE,
                stderr=None if stderr else subprocess.PIPE)
        except subprocess.CalledProcessError as exc:
            raise ShellError(exc=exc) from exc
        else:
            return res.stdout.decode(locale.getpreferredencoding(), 'decode_errors').strip() if res.stdout else ""

    @classmethod
    def _get_curr_info(cls, package):
        cmd = compose_cmd(['dpkg', '-s', package])
        return cls.run_shell_command(cmd)

    @classmethod
    def _parse_deps(cls, deps: str):
        ret = []
        deps = deps.replace("|", ",").replace("(", "").replace(")", "")
        pkgs = deps.split(",")
        for pkg in pkgs:
            if not pkg:
                continue
            name_ver = pkg.strip(" ").split(" ", maxsplit=1)
            name = name_ver[0].strip(" \n")
            if len(name_ver) == 1:
                ret.append(name)
            else:
                ver = name_ver[1].strip("()<>= \n")
                # TODO generally wrong logic since it replaces any
                # fuzzy (>=, < etc.) constraints with equality
                ret.append("{}={}".format(name, ver))
        return ret

    @classmethod
    def _pkgs_dedup(cls, deps):
        ret = []
        processed = set()
        for d in deps:
            name_ver = d.split("=", maxsplit=1)
            if name_ver[0] not in processed:
                ret.append(d)
            processed.add(name_ver[0])
        return ret

    @classmethod
    def _parse_version_deps_from_pkg_mgr_output(
            cls,
            output: str,
            upstream_cls: Type[Union[SourceVersion, None]] = None
    ):
        out_lines = output.split("\n")
        ver = None
        ext_deps = []
        num_pkgs = 0
        for ln in out_lines:
            act_line = ln.strip(" \n")
            if act_line.startswith("Version:"):
                # this method might be used for the dependnecy tree resolving
                # when 'output' includes data for multiple packages,
                # version info here doesn't make any sense in such a case
                num_pkgs += 1
                ver = (None if num_pkgs > 1 else
                       act_line.split(":", maxsplit=1)[1].strip(" \n"))
            if act_line.startswith("Depends:"):
                ext_deps += cls._parse_deps(act_line.split(":", maxsplit=1)[1].strip(" \n"))

        if ver and upstream_cls:
            try:
                ver = DebianVersion(ver, upstream_cls=upstream_cls)
            except InvalidVersionError as exc:
                logger.warning(
                    "Failed to parse debian version {}: {}"
                    .format(ver, exc)
                )
                ver = None
                ext_deps = []
        else:
            ver = None

        return ver, cls._pkgs_dedup(ext_deps)

    @classmethod
    def curr_pkg_info(cls, pkg_name: str) -> Tuple[PackageVersion, List]:
        package_info = cls._get_curr_info(pkg_name)
        return cls._parse_version_deps_from_pkg_mgr_output(
            package_info, upstream_cls=src_version_cls(pkg_name))

    @classmethod
    def get_latest_pkg_version(
            cls,
            pkg_name: str,
            upstream: SourceVersion = None,
            update_cache: bool = True) -> PackageVersion:

        upstream_cls = src_version_cls(pkg_name)

        if upstream and not isinstance(upstream, upstream_cls):
            raise TypeError(
                "'upstream' should be instance of {}, got {}"
                .format(upstream_cls, type(upstream))
            )

        if update_cache:
            cls.update_package_cache()

        try:
            cmd = compose_cmd(
                ['apt-cache', 'show', pkg_name, '|', 'grep', '-E', "'^Version: '"]
            )
            output = cls.run_shell_script_extended(cmd).strip()
        except ShellError:
            # will fail if either package not found or grep returns nothing
            # the latter is unexpected and treated as no-data as well
            logger.info(
                "no data for package '{}' found".format(pkg_name)
            )
        else:
            if output:
                versions = []

                for v in output.split('\n'):
                    try:
                        dv = DebianVersion(v.split()[1], upstream_cls=upstream_cls)
                    except InvalidVersionError as exc:
                        logger.warning(
                            "ignoring invalid version from output {} for upstream class {}: {}"
                            .format(v.split()[1], upstream_cls, exc)
                        )
                    else:
                        if not upstream or (dv.upstream == upstream):
                            versions.append(dv)

                try:
                    return sorted(versions)[-1]
                except IndexError:
                    pass
                except ShellError:
                    logger.warning(
                        "version comparison failed unexpectedly for versions: {}"
                        .format(versions)
                    )

        return None

    @classmethod
    def _get_info_from_package_manager(cls, *package):
        cmd_arg = " ".join(list(package))
        cmd = compose_cmd(['apt-cache', 'show', cmd_arg])
        return cls.run_shell_command(cmd)

    @classmethod
    def update_package_cache(cls):
        cmd = compose_cmd(['apt', 'update'])
        cls.run_shell_script(cmd)

    @classmethod
    def get_deps_tree(cls, *package, depth=0):
        ret = list(set(package))
        if depth < MAX_DEPS_DEPTH:
            package_info = cls._get_info_from_package_manager(*ret)
            _, deps = cls._parse_version_deps_from_pkg_mgr_output(package_info)
            deps_deps = []
            deps = list(set(deps) - set(ret))
            deps_deps.append(cls.get_deps_tree(*deps, depth=depth + 1))

            ret.append(deps_deps)
        return ret

    @classmethod
    def get_deps_tree_filtered(cls, *package, hold_list=[], depth=0, deps_map={}):
        ret = list(set(package))
        if depth < MAX_DEPS_DEPTH:
            package_info = cls._get_info_from_package_manager(*ret)
            _, deps = cls._parse_version_deps_from_pkg_mgr_output(package_info)
            # Make deps list unique
            deps = list(set(deps + ret))
            for d in deps:
                if "=" in d:
                    p, v = d.split("=")
                else:
                    p = d
                    v = ''
                if p in hold_list and \
                        (p not in deps_map or deps_map[p] == ''):
                    deps_map[p] = v
            cls.get_deps_tree_filtered(*deps, hold_list=hold_list, depth=depth + 1, deps_map=deps_map)

        return ["{}={}".format(p, v) if v != '' else p for p, v in deps_map.items()]

    @classmethod
    def dep_tree_traverse(cls, dep_tree, deps_so_far):
        if isinstance(dep_tree, str) and dep_tree not in deps_so_far:
            deps_so_far.append(dep_tree)
        elif isinstance(dep_tree, list) and dep_tree:
            for d in reversed(dep_tree):
                cls.dep_tree_traverse(d, deps_so_far)

    @classmethod
    def get_sys_holds(cls):
        if shutil.which("apt-mark"):
            cmd = compose_cmd(['apt-mark', 'showhold'])
            ret = cls.run_shell_command(cmd)

            hlds = ret.strip().split("\n")
            return [h for h in hlds if h]
        else:
            logger.info('apt-mark not found. Assume holds is empty.')
            return []

    @classmethod
    def hold_packages(cls, packages):
        if shutil.which("apt-mark"):
            packages_to_hold = ' '.join(packages)
            cmd = compose_cmd(['apt-mark', 'hold', packages_to_hold])
            cls.run_shell_script(cmd)
            logger.info('Successfully put {} packages on hold'.format(packages_to_hold))
        else:
            logger.info('Skipping packages holding')



import json

import pytest

from indy_common.authorize.auth_actions import AuthActionEdit, EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.constants import AUTH_RULES, RULES
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.helper import generate_auth_rule
from indy_node.test.helper import sdk_send_and_check_auth_rules_request
from plenum.common.constants import TXN_TYPE
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import sdk_gen_request
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from indy_common.authorize import auth_map


class AuthRulesTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)

    def prepare(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=None)
        self.default_auth_rule = self.get_default_auth_rules()
        self.changed_auth_rule = self.get_changed_auth_rules()

    def run(self):
        # Step 1. Check default auth rule
        sdk_send_and_check_auth_rules_request(self.looper, self.sdk_pool_handle, self.trustee_wallet)
        with pytest.raises(RequestRejectedException):
            sdk_send_and_check_auth_rules_request(self.looper, self.sdk_pool_handle, self.new_default_wallet)

        # Step 2. Change auth rule
        self.send_and_check(self.changed_auth_rule, self.trustee_wallet)

        # Step 3. Check, that we cannot send txn the old way
        with pytest.raises(RequestRejectedException):
            sdk_send_and_check_auth_rules_request(self.looper, self.sdk_pool_handle, self.trustee_wallet)

        # Step 4. Check, that new auth rule is used
        sdk_send_and_check_auth_rules_request(self.looper, self.sdk_pool_handle, self.new_default_wallet)

        # Step 5. Return default auth rule
        self.send_and_check(self.default_auth_rule, self.new_default_wallet)

        # Step 6. Check, that default auth rule works
        sdk_send_and_check_auth_rules_request(self.looper, self.sdk_pool_handle, self.trustee_wallet)
        with pytest.raises(RequestRejectedException):
            sdk_send_and_check_auth_rules_request(self.looper, self.sdk_pool_handle, self.new_default_wallet)

    def result(self):
        pass

    def get_default_auth_rules(self):
        action = AuthActionEdit(txn_type=AUTH_RULES,
                                field='*',
                                old_value='*',
                                new_value='*')
        constraint = auth_map.auth_map.get(action.get_action_id())
        rules = [generate_auth_rule(auth_action=EDIT_PREFIX,
                                    auth_type=AUTH_RULES,
                                    field='*',
                                    old_value='*',
                                    new_value='*',
                                    constraint=constraint.as_dict)]
        operation = {RULES: rules,
                     TXN_TYPE: AUTH_RULES}
        request = sdk_gen_request(operation, identifier=self.new_default_wallet[1])
        return json.dumps(request.as_dict)

    def get_changed_auth_rules(self):
        constraint = AuthConstraint(role=None,
                                    sig_count=1,
                                    need_to_be_owner=False)
        rules = [generate_auth_rule(auth_action=EDIT_PREFIX,
                                    auth_type=AUTH_RULES,
                                    field='*',
                                    old_value='*',
                                    new_value='*',
                                    constraint=constraint.as_dict)]
        operation = {RULES: rules,
                     TXN_TYPE: AUTH_RULES}
        request = sdk_gen_request(operation, identifier=self.trustee_wallet[1])
        return json.dumps(request.as_dict)



from copy import deepcopy

from indy_common.constants import CANCEL, \
    ACTION, SCHEDULE, JUSTIFICATION
from indy_node.test import waits
from indy_node.test.upgrade.helper import checkNoUpgradeScheduled, sdk_send_upgrade
from stp_core.loop.eventually import eventually

whitelist = ['Failed to upgrade node']


def testTrustyCancelsUpgrade(validUpgradeSent, looper, nodeSet, sdk_pool_handle,
                             sdk_wallet_trustee, validUpgrade):
    validUpgradeCopy = deepcopy(validUpgrade)
    validUpgradeCopy[ACTION] = CANCEL
    validUpgradeCopy[JUSTIFICATION] = '"never gonna give you one"'

    validUpgradeCopy.pop(SCHEDULE, None)
    sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_trustee, validUpgradeCopy)

    looper.run(eventually(checkNoUpgradeScheduled, nodeSet, retryWait=1,
                          timeout=waits.expectedNoUpgradeScheduled()))



import json

import pytest
from indy.ledger import build_get_schema_request, parse_get_schema_response

from indy_common.authorize import auth_map
from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import CLAIM_DEF
from indy_node.test.api.helper import sdk_write_schema
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.claim_def.test_send_claim_def import sdk_send_claim_def
from plenum.common.constants import DATA
from plenum.common.exceptions import RequestRejectedException
from plenum.common.types import OPERATION
from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_req, sdk_get_reply, \
    sdk_sign_and_submit_op
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.helper import build_auth_rule_request_json


def get_schema_json(looper, sdk_pool_handle, sdk_wallet_trustee):
    wallet_handle, identifier = sdk_wallet_trustee
    schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_trustee)
    schema_id = json.loads(schema_json)['id']

    request = looper.loop.run_until_complete(build_get_schema_request(identifier, schema_id))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request))[1]
    _, schema_json = looper.loop.run_until_complete(parse_get_schema_response(json.dumps(reply)))
    return schema_json


class AddClaimDefTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def run(self):
        schema_json = get_schema_json(self.looper, self.sdk_pool_handle, self.trustee_wallet)

        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            sdk_send_claim_def(self.looper, self.sdk_pool_handle, self.trustee_wallet, 'tag1', schema_json)

        # Step 3. Check, that new auth rule is used
        sdk_send_claim_def(self.looper, self.sdk_pool_handle, self.new_default_wallet, 'tag2', schema_json)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        sdk_send_claim_def(self.looper, self.sdk_pool_handle, self.trustee_wallet, 'tag3', schema_json)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=CLAIM_DEF,
            field='*',
            new_value='*',
            constraint=constraint.as_dict
        )


class EditClaimDefTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)

        self.add_claim_def = AddClaimDefTest(env, auth_map.add_claim_def.get_action_id())

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def do_edit_claim_def(self, wallet, tag, schema_json):
        reply = sdk_send_claim_def(self.looper, self.sdk_pool_handle, wallet, tag, schema_json)
        req = reply[0][0]
        req[OPERATION][DATA]['primary']['n'] = 'abc'
        resp = sdk_sign_and_submit_op(self.looper, self.sdk_pool_handle, wallet, op=req[OPERATION])
        sdk_get_and_check_replies(self.looper, [resp])

    def run(self):
        schema_json = get_schema_json(self.looper, self.sdk_pool_handle, self.trustee_wallet)
        # Step 1. Change auth rule
        self.send_and_check(self.add_claim_def.get_changed_auth_rule(), self.trustee_wallet)
        self.send_and_check(self.changed_auth_rule, self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            self.do_edit_claim_def(self.trustee_wallet, 'tag_1', schema_json)

        # Step 3, Check, that new auth rule is used
        self.do_edit_claim_def(self.new_default_wallet, 'tag_2', schema_json)

        # Step 4. Return default auth rule
        self.send_and_check(self.add_claim_def.get_default_auth_rule(), self.trustee_wallet)
        self.send_and_check(self.default_auth_rule, self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.do_edit_claim_def(self.trustee_wallet, 'tag_3', schema_json)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=EDIT_PREFIX,
            auth_type=CLAIM_DEF,
            field='*',
            old_value='*',
            new_value='*',
            constraint=constraint.as_dict
        )



import json

import pytest

from indy_common.constants import JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, RS_ID, GET_RICH_SCHEMA_OBJECT_BY_ID, \
    GET_RICH_SCHEMA_OBJECT_BY_METADATA, RS_NAME, RS_VERSION, RS_TYPE
from indy_node.test.api.helper import sdk_build_rich_schema_request, sdk_write_rich_schema_object_and_check
from indy_node.test.helper import rich_schemas_enabled_scope
from indy_node.test.rich_schema.templates import W3C_BASE_CONTEXT
from indy_node.test.rich_schema.test_send_get_rich_schema_obj import PARAMS
from indy_node.test.state_proof.helper import sdk_submit_operation_and_get_result
from plenum.common.constants import TXN_TYPE
from plenum.common.exceptions import RequestNackedException
from plenum.common.util import randomString
from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_and_check_replies


@pytest.fixture(scope='module')
def write_rich_schema(looper, sdk_pool_handle, sdk_wallet_endorser, tconf):
    with rich_schemas_enabled_scope(tconf):
        for txn_type, rs_type, content, rs_id, rs_name, rs_version in PARAMS:
            sdk_write_rich_schema_object_and_check(looper, sdk_wallet_endorser, sdk_pool_handle,
                                                   txn_type=txn_type, rs_id=rs_id, rs_name=rs_name,
                                                   rs_version=rs_version, rs_type=rs_type, rs_content=content)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id',
                         [(JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, W3C_BASE_CONTEXT, randomString())])
def test_send_rich_schema_obj_disabled_by_default(looper, sdk_pool_handle, sdk_wallet_endorser, txn_type, rs_type,
                                                  content, rs_id):
    request = sdk_build_rich_schema_request(looper, sdk_wallet_endorser,
                                            txn_type, rs_id=rs_id, rs_name=randomString(),
                                            rs_version='1.0', rs_type=rs_type,
                                            rs_content=json.dumps(content))

    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_endorser, request)

    with pytest.raises(RequestNackedException, match='RichSchema transactions are disabled'):
        sdk_get_and_check_replies(looper, [req])


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_send_get_rich_schema_obj_by_id_disabled_by_default(looper, sdk_pool_handle, sdk_wallet_endorser, txn_type,
                                                            rs_type, content, rs_id, rs_name, rs_version,
                                                            write_rich_schema):
    get_rich_schema_by_id_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_ID,
        RS_ID: rs_id,
    }

    with pytest.raises(RequestNackedException, match='RichSchema queries are disabled'):
        sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                            sdk_wallet_endorser,
                                            get_rich_schema_by_id_operation)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_send_get_rich_schema_obj_by_metadata_disabled_by_default(looper, sdk_pool_handle, sdk_wallet_endorser,
                                                                  txn_type, rs_type, content, rs_id, rs_name,
                                                                  rs_version, write_rich_schema):
    get_rich_schema_by_metadata_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_METADATA,
        RS_NAME: rs_name,
        RS_VERSION: rs_version,
        RS_TYPE: rs_type
    }

    with pytest.raises(RequestNackedException, match='RichSchema queries are disabled'):
        sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                            sdk_wallet_endorser,
                                            get_rich_schema_by_metadata_operation)



from typing import Union
from datetime import datetime

from common.version import SourceVersion

from indy_common.version import src_version_cls
from indy_common.constants import APP_NAME
from indy_node.server.action_log import ActionLogData, ActionLogEvents, ActionLog


class UpgradeLogData(ActionLogData):
    _items = ActionLogData._items + ['version', 'upgrade_id', 'pkg_name']

    def __init__(
            self,
            when: Union[datetime, str],
            version: Union[SourceVersion, str],
            upgrade_id: str,
            pkg_name: str = APP_NAME
    ):
        super().__init__(when)

        if isinstance(version, str):
            version = src_version_cls(pkg_name)(version)
        if not isinstance(version, SourceVersion):
            raise TypeError(
                "'version' should be 'SourceVersion' or 'str', got: {}"
                .format(type(version))
            )

        self.version = version
        self.upgrade_id = upgrade_id
        self.pkg_name = pkg_name


class UpgradeLog(ActionLog):

    Events = ActionLogEvents

    """
    Append-only event log of upgrade event
    """
    def __init__(self, file_path: str):
        super().__init__(
            file_path,
            data_class=UpgradeLogData,
            event_types=ActionLogEvents,
            delimiter='\t'
        )

    def append_scheduled(self, data: UpgradeLogData):
        super().append_scheduled(data)

    def append_started(self, data: UpgradeLogData):
        super().append_started(data)

    def append_succeeded(self, data: UpgradeLogData):
        super().append_succeeded(data)

    def append_failed(self, data: UpgradeLogData):
        super().append_failed(data)

    def append_cancelled(self, data: UpgradeLogData):
        super().append_cancelled(data)



from plenum.common.constants import TARGET_NYM, TXN_TYPE, NYM, ROLE, VERKEY, \
    CURRENT_PROTOCOL_VERSION
from plenum.common.signer_did import DidIdentity
from stp_core.types import Identifier
from indy_common.auth import Authoriser

from indy_common.generates_request import GeneratesRequest
from indy_common.constants import GET_NYM, NULL
from indy_common.types import Request


class Identity(GeneratesRequest):
    def __init__(self,
                 identifier: Identifier,
                 endorser: Identifier = None,
                 verkey=None,
                 role=None,
                 last_synced=None,
                 seq_no=None):
        """

        :param identifier:
        :param endorser:
        :param verkey:
        :param role: If role is explicitly passed as `null` then in the request
         to ledger, `role` key would be sent as None which would stop the
         Identity's ability to do any privileged actions. If role is not passed,
          `role` key will not be included in the request to the ledger
        :param last_synced:
        :param seq_no:
        """

        self.identity = DidIdentity(identifier, verkey=verkey)
        self.endorser = endorser

        # if role and role not in (ENDORSER, STEWARD):
        if not Authoriser.isValidRole(self.correctRole(role)):
            raise AttributeError("Invalid role {}".format(role))
        self._role = role

        # timestamp for when the ledger was last checked for key replacement or
        # revocation
        self.last_synced = last_synced

        # sequence number of the latest key management transaction for this
        # identifier
        self.seqNo = seq_no

    @property
    def identifier(self):
        return self.identity.identifier

    @property
    def verkey(self):
        return self.identity.verkey

    @verkey.setter
    def verkey(self, new_val):
        identifier = self.identifier
        self.identity = DidIdentity(identifier, verkey=new_val)

    @staticmethod
    def correctRole(role):
        return None if role == NULL else role

    @property
    def role(self):
        return self.correctRole(self._role)

    @role.setter
    def role(self, role):
        if not Authoriser.isValidRole(self.correctRole(role)):
            raise AttributeError("Invalid role {}".format(role))
        self._role = role

    def _op(self):
        op = {
            TXN_TYPE: NYM,
            TARGET_NYM: self.identity.identifier
        }
        if self.identity.verkey is not None:
            op[VERKEY] = None if self.identity.verkey == '' else self.identity.verkey
        if self._role:
            op[ROLE] = self.role
        return op

    def ledgerRequest(self):
        if not self.seqNo:
            assert self.identity.identifier is not None
            return Request(identifier=self.endorser,
                           operation=self._op(),
                           protocolVersion=CURRENT_PROTOCOL_VERSION)

    def _opForGet(self):
        return {
            TARGET_NYM: self.identity.identifier,
            TXN_TYPE: GET_NYM,
        }

    def getRequest(self, requestAuthor: Identifier):
        if not self.seqNo:
            return Request(identifier=requestAuthor,
                           operation=self._opForGet(),
                           protocolVersion=CURRENT_PROTOCOL_VERSION)



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import List
from indy_common.authorize.helper import get_named_role
from indy_common.constants import NETWORK_MONITOR, ENDORSER
from plenum.common.constants import STEWARD, TRUSTEE

CONSTRAINT_ID = "constraint_id"
AUTH_CONSTRAINTS = "auth_constraints"
ROLE = "role"
METADATA = "metadata"
SIG_COUNT = "sig_count"
NEED_TO_BE_OWNER = "need_to_be_owner"
OFF_LEDGER_SIGNATURE = "off_ledger_signature"

IDENTITY_OWNER = None

accepted_roles = [IDENTITY_OWNER, NETWORK_MONITOR, ENDORSER, STEWARD, TRUSTEE, '*']


class ConstraintEnum(Enum):
    def __str__(self):
        return self.name


class ConstraintsEnum:
    ROLE_CONSTRAINT_ID = 'ROLE'
    AND_CONSTRAINT_ID = 'AND'
    OR_CONSTRAINT_ID = 'OR'
    FORBIDDEN_CONSTRAINT_ID = 'FORBIDDEN'

    @staticmethod
    def values():
        return [ConstraintsEnum.ROLE_CONSTRAINT_ID,
                ConstraintsEnum.AND_CONSTRAINT_ID,
                ConstraintsEnum.OR_CONSTRAINT_ID,
                ConstraintsEnum.FORBIDDEN_CONSTRAINT_ID]


class AbstractAuthConstraint(metaclass=ABCMeta):
    def __init__(self):
        self.constraint_id = ''

    @property
    def as_dict(self):
        raise NotImplementedError()

    def __str__(self):
        return str(self)

    def __eq__(self, other):
        if self.as_dict != other.as_dict:
            return False
        return True

    def set_metadata(self, metadata: dict):
        raise NotImplementedError()

    @staticmethod
    def from_dict(as_dict):
        raise NotImplementedError()


class AuthConstraintForbidden(AbstractAuthConstraint):

    def __init__(self):
        self.constraint_id = ConstraintsEnum.FORBIDDEN_CONSTRAINT_ID

    @property
    def as_dict(self):
        return {CONSTRAINT_ID: self.constraint_id}

    def __str__(self):
        return "The action is forbidden"

    def set_metadata(self, metadata: dict):
        pass

    @staticmethod
    def from_dict(as_dict):
        return AuthConstraintForbidden()


class AuthConstraint(AbstractAuthConstraint):
    def __init__(self, role, sig_count,
                 need_to_be_owner=False,
                 off_ledger_signature=None,
                 metadata={}):
        self._role_validation(role, off_ledger_signature)
        self.role = role
        self.sig_count = sig_count
        self.need_to_be_owner = need_to_be_owner
        self.off_ledger_signature = off_ledger_signature
        self.metadata = metadata
        self.constraint_id = ConstraintsEnum.ROLE_CONSTRAINT_ID

    @property
    def as_dict(self):
        constraint = {
            CONSTRAINT_ID: self.constraint_id,
            ROLE: self.role,
            SIG_COUNT: self.sig_count,
            NEED_TO_BE_OWNER: self.need_to_be_owner,
            METADATA: self.metadata
        }
        if self.off_ledger_signature is not None:
            constraint[OFF_LEDGER_SIGNATURE] = self.off_ledger_signature
        return constraint

    @staticmethod
    def _role_validation(role, off_ledger_signature):
        if role not in accepted_roles:
            raise ValueError("Role {} is not acceptable".format(role))
        if off_ledger_signature and role != "*":
            raise ValueError("'off_ledger_signature' can be set to True only if any role is accepted (role='*'). "
                             "Got {} role instead.".format(role))

    def __str__(self):
        role = get_named_role(self.role) if self.role != '*' else 'ALL'
        error_msg = ""

        if self.off_ledger_signature and self.sig_count > 1:
            error_msg = "{} signatures of any role (off-ledger included) are required".format(self.sig_count)
        elif self.off_ledger_signature and self.sig_count == 1:
            error_msg = "1 signature of any role (off-ledger included) is required"

        elif role != 'ALL' and self.need_to_be_owner and self.sig_count > 1:
            error_msg = "{} {} signatures are required and needs to be owner".format(self.sig_count, role)
        elif role != 'ALL' and not self.need_to_be_owner and self.sig_count > 1:
            error_msg = "{} {} signatures are required".format(self.sig_count, role)
        elif role != 'ALL' and not self.need_to_be_owner and self.sig_count == 1:
            error_msg = "1 {} signature is required".format(role)
        elif role != 'ALL' and self.need_to_be_owner and self.sig_count == 1:
            error_msg = "1 {} signature is required and needs to be owner".format(role)

        elif role == "ALL" and self.need_to_be_owner and self.sig_count == 1:
            error_msg = "1 signature of any role is required and needs to be owner"
        elif role == 'ALL' and not self.need_to_be_owner and self.sig_count == 1:
            error_msg = "1 signature of any role is required {}".format(role)
        elif role == 'ALL' and not self.need_to_be_owner and self.sig_count > 1:
            error_msg = "{} signatures of any role are required".format(self.sig_count)
        elif role == "ALL" and self.need_to_be_owner and self.sig_count > 1:
            error_msg = "{} signatures of any role are required and needs to be owner".format(self.sig_count)

        metadata_str = self._metadata_str()
        if metadata_str:
            return "{} {}".format(error_msg, metadata_str)
        return error_msg

    def _metadata_str(self):
        if not self.metadata:
            return ""
        res = " ".join([str(item) for values in self.metadata.items() for item in values])
        return "with additional metadata {}".format(res)

    @staticmethod
    def from_dict(as_dict):
        return AuthConstraint(role=as_dict[ROLE], sig_count=as_dict[SIG_COUNT],
                              need_to_be_owner=as_dict.get(NEED_TO_BE_OWNER, False),
                              off_ledger_signature=as_dict.get(OFF_LEDGER_SIGNATURE, None),
                              metadata=as_dict.get(METADATA, {}))

    def set_metadata(self, metadata: dict):
        self.metadata = metadata


class AuthConstraintAnd(AbstractAuthConstraint):
    def __init__(self, auth_constraints: List[AbstractAuthConstraint]):
        self.auth_constraints = auth_constraints
        self.constraint_id = ConstraintsEnum.AND_CONSTRAINT_ID

    @property
    def as_dict(self):
        return {
            CONSTRAINT_ID: self.constraint_id,
            AUTH_CONSTRAINTS: [c.as_dict for c in self.auth_constraints]
        }

    def __str__(self):
        return " AND ".join([str(ac) for ac in self.auth_constraints])

    @staticmethod
    def from_dict(as_dict):
        auth_constraints = []
        for input_constraint in as_dict[AUTH_CONSTRAINTS]:
            dict_constraint = dict(input_constraint)
            constraint_id = dict_constraint.pop(CONSTRAINT_ID)
            constraint_cls = constraint_to_class_map.get(constraint_id)
            auth_constraints.append(constraint_cls.from_dict(dict_constraint))
        as_dict[AUTH_CONSTRAINTS] = auth_constraints

        return AuthConstraintAnd(**as_dict)

    def set_metadata(self, metadata: dict):
        for constraint in self.auth_constraints:
            constraint.set_metadata(metadata)


class AuthConstraintOr(AbstractAuthConstraint):
    def __init__(self, auth_constraints: List[AbstractAuthConstraint]):
        self.auth_constraints = auth_constraints
        self.constraint_id = ConstraintsEnum.OR_CONSTRAINT_ID

    @property
    def as_dict(self):
        return {
            CONSTRAINT_ID: self.constraint_id,
            AUTH_CONSTRAINTS: [c.as_dict for c in self.auth_constraints]
        }

    def __str__(self):
        return " OR ".join([str(ac) for ac in self.auth_constraints])

    @staticmethod
    def from_dict(as_dict):
        auth_constraints = []
        for input_constraint in as_dict[AUTH_CONSTRAINTS]:
            dict_constraint = dict(input_constraint)
            constraint_id = dict_constraint.pop(CONSTRAINT_ID)
            if constraint_id is None:
                raise KeyError('There is no "constraint_id" field in deserialised dict: {}'.format(as_dict))
            constraint_cls = constraint_to_class_map.get(constraint_id)
            auth_constraints.append(constraint_cls.from_dict(dict_constraint))
        as_dict[AUTH_CONSTRAINTS] = auth_constraints

        return AuthConstraintOr(**as_dict)

    def set_metadata(self, metadata: dict):
        for constraint in self.auth_constraints:
            constraint.set_metadata(metadata)


class ConstraintCreator:
    @staticmethod
    def create_constraint(input_dict):
        as_dict = dict(input_dict)
        constraint_id = as_dict.pop(CONSTRAINT_ID)
        if constraint_id is None:
            raise KeyError('There is no "constraint_id" field in deserialised dict: {}'.format(as_dict))

        constraint_cls = constraint_to_class_map.get(constraint_id)
        return constraint_cls.from_dict(as_dict)


class AbstractConstraintSerializer(metaclass=ABCMeta):
    def __init__(self, serializer):
        self.serializer = serializer

    @abstractmethod
    def serialize(self, constraint: AbstractAuthConstraint) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def deserialize(self, serialized_str: bytes) -> AbstractAuthConstraint:
        raise NotImplementedError()


class ConstraintsSerializer(AbstractConstraintSerializer):
    def serialize(self, constraint: AbstractAuthConstraint) -> bytes:
        return self.serializer.serialize(constraint.as_dict if constraint else {})

    def deserialize(self, serialized_str: bytes) -> AbstractAuthConstraint:
        as_dict = self.serializer.deserialize(serialized_str)
        return ConstraintCreator.create_constraint(as_dict)


class AbstractAuthConstraintParser(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def is_accepted(constraint_results: List):
        raise NotImplementedError()


class AuthConstraintParserOr(AbstractAuthConstraintParser):
    @staticmethod
    def is_accepted(constraint_results: List):
        return any(constraint_results)


class AuthConstraintParserAnd(AbstractAuthConstraintParser):
    @staticmethod
    def is_accepted(constraint_results: List):
        return all(constraint_results)


constraint_to_class_map = {
    ConstraintsEnum.ROLE_CONSTRAINT_ID: AuthConstraint,
    ConstraintsEnum.AND_CONSTRAINT_ID: AuthConstraintAnd,
    ConstraintsEnum.OR_CONSTRAINT_ID: AuthConstraintOr,
    ConstraintsEnum.FORBIDDEN_CONSTRAINT_ID: AuthConstraintForbidden,
}



import functools

import pytest

from indy_common.constants import POOL_UPGRADE
from indy_node.server.node import Node
from plenum.common.constants import FORCE, TXN_TYPE
from plenum.test.helper import sdk_gen_request
from plenum.test.primary_selection.test_view_changer_primary_selection import FakeNode


@pytest.fixture(scope='function')
def fake_node(tdir, tconf):
    node = FakeNode(tdir, config=tconf)
    node.unpackClientMsg = functools.partial(Node.unpackClientMsg, node)
    node.is_request_need_quorum = functools.partial(Node.is_request_need_quorum, node)
    node.master_replica._consensus_data.waiting_for_new_view = True
    fake_node.postToClientInBox = lambda a, b: None
    return node


def test_client_force_request_not_discard_in_view_change_with_dict(fake_node):
    sender = "frm"
    msg = sdk_gen_request({TXN_TYPE: POOL_UPGRADE, FORCE: True}).as_dict

    def post_to_client_in_box(received_msg, received_frm):
        assert received_frm == sender
        assert received_msg == msg
    fake_node.postToClientInBox = post_to_client_in_box

    def discard(received_msg, reason, logMethod, cliOutput):
        assert False, "Message {} was discard with '{}'".format(received_msg, reason)
    fake_node.discard = discard

    fake_node.unpackClientMsg(msg, sender)
